/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package drjava.util.i18n;

import org.apache.bcel.classfile.*;
import org.apache.bcel.util.ClassPath;
import org.apache.bcel.util.SyntheticRepository;
import org.apache.bcel.generic.*;

import java.io.*;
import java.util.*;

public class BytecodeI18n {
  private static final String AREA_TRANSLATE = "translate";
  private static final String AREA_IGNORE = "ignore";
  private static final String AREA_OBSOLETE = "obsolete";
  private static final String TRANSLATION_CLASS = "drjava.superversion.I18n";
  private static final String TRANSLATION_METHOD = "translate";
  private static final String TRANSLATION_SIGNATURE = "(Ljava/lang/String;)Ljava/lang/String;";

  private static Set<String> strings = new TreeSet<String>();
  private static int rewrites;
  private static int rewritesUndone;

  public static void main(String[] args) throws ClassNotFoundException, IOException, TargetLostException {
    String cp = null, outputDir = null;
    boolean rewrite = false;
    for (Iterator<String> i = Arrays.asList(args).iterator(); i.hasNext(); ) {
      String arg = i.next();
      if (arg.equals("-cp"))
        cp = i.next();
      else if (arg.equals("-d"))
        outputDir = i.next();
      else if (arg.equals("-rewrite"))
        rewrite = true;
    }
    if (cp == null)
      throw new RuntimeException("Missing argument -cp");
    if (outputDir == null)
      throw new RuntimeException("Missing argument -outputDir");
    System.out.println("classpath: " + cp + ", outputDir: " + outputDir);
    List<String> classNames = ClasspathScanner.findAllClasses(new File(cp));
    System.out.println(classNames.size() + " classes found");
    for (ListIterator<String> i = classNames.listIterator(); i.hasNext(); )
      if (!includeClass(i.next()))
        i.remove();
    System.out.println(classNames.size() + " classes relevant");

    SyntheticRepository repository = SyntheticRepository.getInstance(new ClassPath(cp));

    for (String className : classNames) {
      //System.out.print(className + " ");
      JavaClass clazz = repository.loadClass(className);
      scanClass(clazz);
      //System.out.println(" - " + strings.size() + " strings");
    }

    System.out.println(strings.size() + " strings found");

    String newArea = "new " + new Date().toString();
    System.out.println("Processing en.properties");
    File enFile = new File(outputDir, "en.properties");
    Map<String, Properties> enMap = readTranslationsFile(enFile);
    enRelocate(enMap);
    addNewStrings(enMap, strings, false, newArea);
    saveTranslationsFile(enMap, enFile);

    for (File file : new File(outputDir).listFiles()) {
      if (file.getName().endsWith(".properties") && file.getName().length() == 13
        && !file.getName().equals("en.properties")) {
        System.out.println("Processing " + file.getName());
        Map<String, Properties> map = readTranslationsFile(file);
        nonEnRelocate(map, enMap);
        if (enMap.containsKey(AREA_TRANSLATE))
          addNewStrings(map, (Collection) enMap.get(AREA_TRANSLATE).keySet(), true, newArea);
        saveTranslationsFile(map, file);
      }
    }

    if (rewrite) {
      System.out.println("Rewriting class files");
      for (String className : classNames) {
        JavaClass clazz = repository.loadClass(className);
        File file = new File(cp, className + ".class");
        rewriteClass(file, clazz, enMap);
      }
      System.out.println("Done. " + rewritesUndone + " rewrites undone, " + rewrites + " new");
    }
  }

  private static void rewriteClass(File file, JavaClass clazz, Map<String, Properties> map) throws IOException, TargetLostException {
    Properties properties = map.get("translate");
    if (properties == null) return;
    ClassGen classGen = new ClassGen(clazz);
    ConstantPoolGen constantPoolGen = classGen.getConstantPool();
    boolean classModified = false;
    Method[] methods = clazz.getMethods();
    for (int iMethod = 0; iMethod < methods.length; iMethod++) {
      Method method = methods[iMethod];
      MethodGen mg = new MethodGen(method, clazz.getClassName(), constantPoolGen);
      InstructionList il = mg.getInstructionList();
      if (il == null)
        continue;
      boolean methodModified = false;

      // remove all translation calls
      repeat: while (true) {
        for (InstructionHandle instructionHandle : il.getInstructionHandles()) {
          Instruction instruction = instructionHandle.getInstruction();
          if (instruction instanceof INVOKESTATIC) {
            INVOKESTATIC inv = (INVOKESTATIC) instruction;
            if (inv.getClassName(constantPoolGen).equals(TRANSLATION_CLASS)
              && inv.getMethodName(constantPoolGen).equals(TRANSLATION_METHOD)
              && inv.getSignature(constantPoolGen).equals(TRANSLATION_SIGNATURE)) {
              il.delete(instructionHandle);
              methodModified = true;
              ++rewritesUndone;
              continue repeat;
            }
          }
        }
        break;
      }

      // insert new translation calls

      for (InstructionHandle instructionHandle : il.getInstructionHandles()) {
        Instruction instruction = instructionHandle.getInstruction();
        if (instruction instanceof LDC) {
          CPInstruction cpInstr = (CPInstruction) instruction;
          Constant constant = constantPoolGen.getConstant(cpInstr.getIndex());
          if (constant instanceof ConstantString) {
            String str = ((ConstantString) constant).getBytes(constantPoolGen.getConstantPool());
            if (properties.containsKey(str)) {
              InstructionList newInstructions = new InstructionList();
              int poolIdx = constantPoolGen.addMethodref(
                TRANSLATION_CLASS, TRANSLATION_METHOD, TRANSLATION_SIGNATURE);
              System.out.println("poolIdx="+poolIdx);
              newInstructions.append(new INVOKESTATIC(poolIdx));
              il.append(instruction, newInstructions);
              ++rewrites;
              methodModified = true;
            }
          }
        }
      }
      if (methodModified) {
        classGen.setMethodAt(mg.getMethod(), iMethod);
        classModified = true;
      }
    }
    if (classModified) {
      System.out.println("Saving " + file);
      classGen.getJavaClass().dump(file);
    }
  }

  private static void enRelocate(Map<String, Properties> map) {
    ArrayList<Map.Entry<String, Properties>> entries = new ArrayList<Map.Entry<String, Properties>>(map.entrySet());
    for (Map.Entry<String, Properties> entry : entries) {
      String area = entry.getKey();
      Properties properties = entry.getValue();
      ArrayList<Map.Entry<String, String>> entries2 =
        new ArrayList<Map.Entry<String, String>>((Set) properties.entrySet());
      for (Map.Entry<String, String> entry2 : entries2) {
        String key = entry2.getKey();
        String value = entry2.getValue();
        if (!strings.contains(key)) {
          properties.remove(key);
          addString(map, key, value, AREA_OBSOLETE);
        } else if (value.equalsIgnoreCase("i") && !area.equals(AREA_IGNORE)) {
          properties.remove(key);
          addString(map, key, value, AREA_IGNORE);
        } else if (value.equalsIgnoreCase("t") && !area.equals(AREA_TRANSLATE)) {
          properties.remove(key);
          addString(map, key, value, AREA_TRANSLATE);
        }
      }
    }
  }

  private static void nonEnRelocate(Map<String, Properties> map, Map<String, Properties> enMap) {
    // move obsolete translations to "obsolete" category
    ArrayList<Map.Entry<String, Properties>> entries = new ArrayList<Map.Entry<String, Properties>>(map.entrySet());
    for (Map.Entry<String, Properties> entry : entries) {
      Properties properties = entry.getValue();
      ArrayList<Map.Entry<String, String>> entries2 =
        new ArrayList<Map.Entry<String, String>>((Set) properties.entrySet());
      for (Map.Entry<String, String> entry2 : entries2) {
        String key = entry2.getKey();
        String value = entry2.getValue();
        if (!strings.contains(key) || hasString(enMap, key, AREA_IGNORE)) {
          properties.remove(key);
          addString(map, key, value, AREA_OBSOLETE);
        }
      }
    }
  }

  private static boolean includeClass(String className) {
    return className.startsWith("drjava/superversion/")
      && className.indexOf("/test") < 0
      && !className.equals(TRANSLATION_CLASS);
  }

  private static void addNewStrings(Map<String, Properties> map,
                                    Collection<String> strings, boolean values,
                                    String area) {
    for (String s : strings) {
      if (!hasString(map, s)) {
        addString(map, s, values ? s : "", area);
      }
    }
  }

  private static void addString(Map<String, Properties> map, String key, String value, String area) {
    if (!map.containsKey(area))
      map.put(area, new Properties());
    map.get(area).put(key, value);
  }

  private static boolean hasString(Map<String, Properties> map, String s, String area) {
    return map.containsKey(area) && map.get(area).containsKey(s);
  }

  private static boolean hasString(Map<String, Properties> map, String s) {
    for (Properties properties : map.values()) {
      if (properties.containsKey(s))
        return true;
    }
    return false;
  }

  private static void saveTranslationsFile(Map<String, Properties> map, File file) throws IOException {
    File newFile = new File(file.getParent(), "new_" + file.getName());
    PrintStream out = new PrintStream(new FileOutputStream(newFile));
    for (Map.Entry<String, Properties> entry : map.entrySet()) {
      out.println("");
      out.println("## " + entry.getKey());
      out.println("");
      ByteArrayOutputStream tempOut = new ByteArrayOutputStream();
      entry.getValue().store(tempOut, "");
      DataInputStream tempIn = new DataInputStream(new StringBufferInputStream(tempOut.toString()));
      String line;
      List<String> lines = new ArrayList<String>();
      while ((line = tempIn.readLine()) != null) {
        if (line.length() != 0 && !line.startsWith("#"))
          lines.add(line);
      }
      Collections.sort(lines);
      for (String s : lines) {
        out.println(s);
      }
    }
    out.close();
    file.delete();
    newFile.renameTo(file);
  }

  private static Map<String,Properties> readTranslationsFile(File file) throws IOException {
    Map<String, Properties> map = new TreeMap<String, Properties>();
    DataInputStream in = new DataInputStream(new FileInputStream(file));
    String line;
    StringBuffer buf = null;
    Properties props = null;
    while ((line = in.readLine()) != null) {
      if (line.startsWith("## ")) {
        if (props != null)
          props.load(new StringBufferInputStream(buf.toString()));
        props = new Properties();
        map.put(line.substring(3), props);
        buf = new StringBuffer();
      } else {
        if (buf != null)
          buf.append(line).append('\n');
      }
    }
    if (props != null)
      props.load(new StringBufferInputStream(buf.toString()));
    in.close();
    return map;
  }

  private static void scanClass(JavaClass clazz) {
    ConstantPool constantPool = clazz.getConstantPool();
    for (int i = 0; i < constantPool.getLength(); i++) {
      Constant constant = constantPool.getConstant(i);
      if (constant instanceof ConstantString) {
        ConstantString cStr = (ConstantString) constant;
        String str = cStr.getBytes(constantPool);
        if (!ignoreString(str)) {
          strings.add(str);
        }
        //System.out.println("idx " + i + ": " + str);
      }
    }
  }

  private static boolean ignoreString(String str) {
    return str.equals("") || str.startsWith("000000");
  }
}