/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.audio.a03;

import prophecy.common.Marker;
import prophecy.common.sound.Sound;

import static java.lang.Math.abs;
import static java.lang.Math.max;
import java.util.ArrayList;
import java.util.List;

public class AudioAnalysis {
  public static double sineSum(Sound sound, double freq) {
    return average(sineMul(sound, freq));
  }

  private static Sound sineMul(Sound sound, double freq) {
    return mul(sound, new Sine(sound.getSampleRate(), sound.getSeconds(), freq));
  }

  public static double cosineSum(Sound sound, double freq) {
    return average(cosineMul(sound, freq));
  }

  private static Sound cosineMul(Sound sound, double freq) {
    return mul(sound, new Cosine(sound.getSampleRate(), sound.getSeconds(), freq));
  }

  public static Sound mul(Sound a, Sound b) {
    float[] samples = new float[Math.min(a.getNumSamples(), b.getNumSamples())];
    for (int i = 0; i < samples.length; i++)
      samples[i] = a.getSample(i)*b.getSample(i);
    return new Sound(a.getSampleRate(), samples);
  }

  public static Sound mul(Sound a, double volume) {
    if (volume == 1.0) return a;
    float[] samples = new float[a.getNumSamples()];
    for (int i = 0; i < samples.length; i++)
      samples[i] = (float) (a.getSample(i)*volume);
    return new Sound(a.getSampleRate(), samples);
  }

  public static double average(Sound sound) {
    double sum = 0;
    for (float s : sound.getSamples()) {
      sum += s;
    }
    return sum / sound.getNumSamples();
  }

  public static double absAverage(Sound sound) {
    double sum = 0;
    for (float s : sound.getSamples()) {
      sum += Math.abs(s);
    }
    return sum / sound.getNumSamples();
  }

  public static List<Sound> clips(Sound sound, double clipSize) {
    List<Sound> clips = new ArrayList<Sound>();
    for (double s = 0; s < sound.getSeconds(); s += clipSize) {
      Marker m1 = new Marker(s), m2 = new Marker(s+clipSize);
      clips.add(sound.clip(m1, m2));
    }
    return clips;
  }

  public static Sound normalize(Sound sound) {
    double max = getMaxAmplitude(sound);
    return mul(sound, max == 0 ? 1.0 : 1.0/max);
  }

  private static double getMaxAmplitude(Sound sound) {
    double max = 0;
    for (double sample : sound.getSamples())
      max = max(max, abs(sample));
    return max;
  }

  public static Sound resample(Sound sound, double speed) {
    float[] samples = new float[(int) Math.floor(sound.getNumSamples()/speed)];
    for (int i = 0; i < samples.length; i++)
      samples[i] = sound.getSample((int) (i*speed));
    return new Sound(sound.getSampleRate(), samples);
  }
}