/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.audio.a06;

import prophecy.common.Prophecy;
import prophecy.audio.a01.*;
import prophecy.audio.a03.*;
import prophecy.common.*;
import static prophecy.common.InspectionUtil.inspect;
import static prophecy.common.SurfaceUtil.show;
import prophecy.common.sound.Sound;
import prophecy.common.sound.SoundSurface;
import prophecy.common.sound.SoundUtil;
import prophecy.common.sound.SoundWithMarkers;

import static java.lang.System.arraycopy;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class A06 {
  private SoundWithMarkers example1;
  private SoundWithMarkers example2;
  public static final String CONTEST_STAGE0 = "a06_stage0";

  public static void main(String[] args) {
    new A06().run();
  }

  public A06() {
    double[] example1markers = { 0.197, 0.922, 2.124, 2.246, 2.847, 4.040, 4.766, 5.967, 6.083, 6.683 };
    example1 = new SoundWithMarkers(
      SoundUtil.loadSound(Prophecy.dataDir() + "audio/flawless.wav"), example1markers);
    example2 = new SoundWithMarkers(
      SoundUtil.loadSound(Prophecy.dataDir() + "audio/flawless2.wav"), shift(example1markers, 0.38-0.197));
  }

  public void run() {
    saveContests();
    //showFunctions(example1);

    //showFreqAnalysis(example1.sound, 50, 75, 100, 200, 400);

    //SurfaceUtil.show("A6", new SoundSurface(example1));
    //testDeafContestant(example1);

    //testClipification(example1, example2);
  }

  private void saveContests() {
    BaseDrumContest contest = new BaseDrumContest(example1);
    contest.setTrainer(new TrainableTrainer<SoundWithMarkers, AudioEventFinder>(Arrays.asList(example1)));
    ContestUtil.saveContest(contest, "a06_stage0");

    contest = new BaseDrumContest(example1);
    contest.setTrainer(new TrainableTrainer<SoundWithMarkers, AudioEventFinder>(Arrays.asList(example1)));
    ContestUtil.saveContest(contest, "a06_stage1");
  }

  private static void showFunctions(SoundWithMarkers example) {
    SoundSurface surface = new SoundSurface(example);

    // show clip ends
    for (Marker marker : example.markers)
      surface.addMarker(marker.plus(0.1));

    surface.addFunction(featureFunction(new AutoScaleFeature<Sound>(new Loudness(), -10, -2), 0.1));
    /*for (double freq : new double[] { 50, 100, 200 }) {
      surface.addFunction(featureFunction(new AutoScaleFeature<Sound>(new Ln<Sound>(new FreqFeature(freq)), -10, -2)));
    }*/
    for (double clipSize : new double[] { 0.01, 0.05, 0.1 }) {
      surface.addFunction(featureFunction(new AutoScaleFeature<Sound>(new Ln<Sound>(new FreqFeature(100)), -10, -2), clipSize));
    }
    show("A6", surface);
  }

  private static SoundFunction featureFunction(final Feature<Sound> feature, final double clipSize) {
    return new SoundFunction() {
      public double getValue(Sound sound, double second) {
        double l = feature.extract(sound.clip(new Marker(second), new Marker(second + clipSize)));
        return l;
      }
    };
  }

  private static double[] shift(double[] markers, double x) {
    double[] newMarkers = new double[markers.length];
    for (int i = 0; i < markers.length; i++) newMarkers[i] = markers[i]+x;
    return newMarkers;
  }

  private static void showFreqAnalysis(Sound sound, double... freqs) {
    List<InspectionNode> nodes = new ArrayList<InspectionNode>();
    nodes.add(new InspectionNode("original", sound));
    for (double freq : freqs) {
      Sine sine = new Sine(sound.getSampleRate(), sound.getSeconds(), freq);
      nodes.add(new InspectionNode("sin " + freq, sine));
      nodes.add(new InspectionNode("mul", AudioAnalysis.mul(sound, sine)));
    }
    inspect(nodes);
  }

  private static void testClipification(SoundWithMarkers example1, SoundWithMarkers example2) {
    Clipifier clipifier = new Clipifier(0.1);
    //InspectionUtil.showInspectionFrame(clipifier.makeClips(example1.sound));
    StandardSubsetContest<Sound> contest = new StandardSubsetContest<Sound>();
    List<SubsetTI<Sound>> trainingItems = clipifier.makeTrainingItems(example1);
    contest.addContestItems(trainingItems);
    contest.setTrainer(new IgnoreUntrainables<Subset<Sound>>(
      new SubsetAutoTrainer<Sound, Trainable<SubsetTI<Sound>>>(contest)));

    List<Feature<Sound>> features = A3.makeFreqFeatures(50, 75, 100, 200, 400);
    features.add(new AutoScaleFeature<Sound>(new Loudness()));
    ContestRun<Subset<Sound>> contestRun = new ExperimentWithFeatures<Sound>(contest, features).run(10);

    List<Object> things = new ArrayList<Object>();
    things.add(new InspectionNode("example1", example1));
    things.add(new InspectionNode("example2", example2));
    things.add(trainingItems);
    things.add(contestRun.getWinner());
    things.add(new SoundDiscriminationSurface(concatSounds(trainingItems), contest,
      clipifier.getClipSize(), contestRun.getWinner()));
    inspect("A6", things);
  }

  private static Sound concatSounds(List<SubsetTI<Sound>> tis) {
    int totalSamples = 0;
    for (SubsetTI<Sound> ti : tis) totalSamples += ti.example.getNumSamples();
    float[] samples = new float[totalSamples];
    int i = 0;
    for (SubsetTI<Sound> ti : tis) {
      arraycopy(ti.example.getSamples(), 0, samples, i, ti.example.getNumSamples());
      i += ti.example.getNumSamples();
    }
    return new Sound(tis.get(0).example.getSampleRate(), samples);
  }

  private static void testDeafContestant(SoundWithMarkers example1) {
    BaseDrumContest contest = new BaseDrumContest(example1);
    contest.setTrainer(new TrainableTrainer<SoundWithMarkers, AudioEventFinder>(Arrays.asList(example1)));
    contest.addContestant(new DeafContestant());
    contest.run();
  }

  public SoundWithMarkers getExample1() {
    return example1;
  }

  public SoundWithMarkers getExample2() {
    return example2;
  }
}