/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.ce.ce01;

import static drjava.util.ObjectUtil.nice;
import drjava.util.Port;
import drjava.util.StringUtil;
import prophecy.m.m02.Typewriter;
import prophecy.m.m09.JReport;
import prophecy.m.m09.Reporter;
import prophecy.t.t08.Bag;
import prophecy.common.CodeQueue;

import javax.swing.*;
import java.util.ArrayList;
import java.util.List;

/** advanced version of t09.Soup */
public class Soup implements Reporter {
  private Bag<Object> ingredients = new Bag<Object>();
  private List<Object> additionBuffer;
  private List<SoupListener> listeners = new ArrayList<SoupListener>();
  private boolean autoStir;
  private boolean autoInitiated;
  //private List<Filter> filters = new ArrayList<Filter>();
  private boolean stirred = true;

  public void add(Object object) {
    ActiveIngredient ingredient = null;
    if (object instanceof ActiveIngredient)
      ingredient = (ActiveIngredient) object;

    if (ingredient != null) {
      ingredient.setSoup(this);
      if (ingredient.isDuplicate()) {
        System.out.println("Dropping duplicate: " + ingredient);
        return;
      }
    }

    if (additionBuffer != null) {
      additionBuffer.add(object);
      return;
    }

    /*for (Filter filter : filters) {
      if (!filter.isAccepted(object))
        return;
    }*/

    ingredients.add(object);
    stirred = false;

    for (SoupListener listener : listeners) {
      listener.ingredientAdded(this, object);
    }

    if (ingredient != null) {
      ingredient.run();
    }

    if (autoStir && !autoInitiated && CodeQueue.get(this) != null) {
      autoInitiated = true;
      CodeQueue.get(this).add(new Runnable() {
        public void run() {
          autoInitiated = false;
          stir();
          sift();
        }
      });
    }
  }

  public void addSoupListener(SoupListener listener) {
    listeners.add(listener);
  }

  public List<Object> getIngredients() {
    return ingredients.getSnapshot();
  }

  public <A> List<A> getIngredients(Class<A> aClass) {
    List<A> result = new ArrayList<A>();
    for (Object ingredient : ingredients.getSnapshot()) {
      if (aClass.isInstance(ingredient))
        result.add((A) ingredient);
    }
    return result;
  }

  public <A> A getIngredient(Class<A> aClass) {
    List<A> list = getIngredients(aClass);
    return list.isEmpty() ? null : list.get(0);
  }

  public void stir() {
    suspendAddition();
    try {
      // newer ingredients make better reaction partners
      for (int i = ingredients.size()-1; i >= 0; i--) {
        Object a = ingredients.get(i);
        if (!(a instanceof ActiveIngredient)) continue;
        ActiveIngredient activeIngredient = (ActiveIngredient) a;
        if (activeIngredient.getReplacement() != null) continue;
        List<PossibleReaction> possibleReactions = activeIngredient.getPossibleReactions();
        if (possibleReactions.isEmpty()) continue;

        // newer ingredients make better reaction partners
        for (int j = ingredients.size()-1; j >= 0; j--) {
          Object b = ingredients.get(j);
          if (hasReplacement(b)) continue;
          for (PossibleReaction possibleReaction : possibleReactions) {
            if (possibleReaction.getBClass().isInstance(b))
              if (possibleReaction.runWith(b))
                System.out.println("Reaction done: " + nice(possibleReaction));
          }
        }
      }
    } finally {
      resumeAddition();
    }

    sift();
    stirred = true;
  }

  private boolean hasReplacement(Object o) {
    return o instanceof ActiveIngredient && ((ActiveIngredient) o).getReplacement() != null;
  }

  private void suspendAddition() {
    additionBuffer = new ArrayList<Object>();
  }

  private void resumeAddition() {
    List<Object> newObjects = additionBuffer;
    additionBuffer = null;
    for (Object object : newObjects) {
      add(object);
    }
  }

  /** remove objects that have a replacement */
  private void sift() {
    suspendAddition();
    try {
      Bag<Object> survivors = new Bag<Object>();
      for (Object ingredient : ingredients.getSnapshot()) {
        if (ingredient instanceof ActiveIngredient) {
          ActiveIngredient activeIngredient = (ActiveIngredient) ingredient;
          Object replacement = activeIngredient.getReplacement();
          if (replacement != null) {
            if (!ingredients.contains(replacement)) {
              //System.out.println("Warning: Replacement " + replacement + " of " + ingredient + " not in soup");
              System.out.println("Adding replacement " + replacement + " of " + ingredient + " to soup");
              add(replacement);
            }
            continue;
          }
        }
        survivors.add(ingredient);
      }
      ingredients = survivors;
    } finally {
      resumeAddition();
    }
  }

  public int size() {
    return ingredients.size();
  }

  public Port<Object> getThreadSafePort() {
    final CodeQueue codeQueue = CodeQueue.get(this);
    if (codeQueue == null)
      throw new RuntimeException("No event queue set");

    return new Port<Object>() {
      public void receive(final Object o) {
        codeQueue.add(new Runnable() {
          public void run() {
            add(o);
          }
        });
      }
    };
  }

  public Typewriter writeReport() {
    Typewriter typewriter = new Typewriter();
    for (Object ingredient : ingredients.getSnapshot()) {
      typewriter.addLine(StringUtil.escapeHtml(nice(ingredient)));
    }
    return typewriter;
  }

  public JComponent visualize() {
    return new JReport(this);
  }

  public boolean getAutoStir() {
    return autoStir;
  }

  public void setAutoStir(boolean autoStir) {
    this.autoStir = autoStir;
  }

  /*public void addFilter(Filter filter) {
    filters.add(filter);
  }*/

  public boolean isStirred() {
    return stirred;
  }

  public void remove(Object object) {
    if (object instanceof ActiveIngredient) {
      ActiveIngredient ingredient = (ActiveIngredient) object;
      ingredient.setSoup(null);
    }

    ingredients.remove(object);

    for (SoupListener listener : listeners) {
      listener.ingredientRemoved(this, object);
    }
  }

  public boolean contains(Object ingredient) {
    return ingredients.contains(ingredient);
  }

  public void assignEventQueue() {
    CodeQueue codeQueue = new CodeQueue();
    codeQueue.startThread();
    codeQueue.attachTo(this);
  }

  public void stopEventThread() {
    CodeQueue codeQueue = CodeQueue.get(this);
    if (codeQueue != null)
      codeQueue.stopThreads();
  }
}