/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.*;
import prophecy.common.Scores;

import javax.swing.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Locale;

public abstract class AbstractContestRun<A> implements Displayable, IContestRun<A>, ContestantReceiver<A> {
  protected Trainer<A> trainer;
  protected boolean persistTrainer = true;
  private ArrayList<A> contestants = new ArrayList<A>();
  private Scores<A> scores;
  protected int rounds = 1;

  public AbstractContestRun() {
  }

  public void addContestant(A contestant) {
    contestants.add(contestant);
  }

  public void addContestants(Collection<? extends A> contestants) {
    this.contestants.addAll(contestants);
  }

  public void runSilently() {
    training();

    for (int i = 0; i < rounds; i++) {
        oneRound();
    }
  }

  public void run() {
    training();

    Locale.setDefault(Locale.ENGLISH);
    ProgressMonitor progressMonitor = new ProgressMonitor(null, "Running compo", "", 0, rounds);
    try {
      for (int i = 0; i < rounds; i++) {
        progressMonitor.setNote("Round " + (i+1) + " of " + rounds);
        oneRound();
        progressMonitor.setProgress(i);
      }
    } finally {
      progressMonitor.close();
    }

    printScores();
  }

  public void training() {
    prepare();
    train();
  }

  protected final void train() {
    for (A contestant : contestants)
      trainContestant(contestant);
  }

  protected final void trainContestant(A contestant) {
    if (trainer != null)
      trainer.train(contestant);
  }

  private void oneRound() {
    List<Double> scores = MultiCoreUtil.parallelMap(contestants, new Function<A, Double>() {
      public Double get(A contestant) {
        System.out.println("  # " + Thread.currentThread().getName() + " evaluating " + contestant);
        return evaluate(contestant);
      }
    });
    for (int i = 0; i < contestants.size(); i++)
      this.scores.add(contestants.get(i), scores.get(i));
  }

  public abstract double evaluate(A contestant);

  protected final void prepare() {
    scores = new Scores<A>();
    scores.setBestPossibleScore(getBestPossibleScore());
    scores.setWorstPossibleScore(getWorstPossibleScore());
    customPreparation();
  }

  protected void customPreparation() {
  }

  private void printScores() {
    display(StdOutDisplay.get());
  }

  public void display(Display display) {
    display.nl().putnl(ObjectUtil.toNiceString(this)).put(scores);
  }

  public void setRounds(int rounds) {
    this.rounds = rounds;
  }

  public abstract Double getBestPossibleScore();

  public abstract Double getWorstPossibleScore();

  public A getWinner() {
    return scores.getWinner();
  }

  public Scores<A> getScores() {
    return scores;
  }

  public Trainer<A> getTrainer() {
    return trainer;
  }

  public void setTrainer(Trainer<A> trainer) {
    setTrainer(trainer, true);
  }

  public void setTrainer(Trainer<A> trainer, boolean persistTrainer) {
    this.trainer = trainer;
    this.persistTrainer = persistTrainer;
  }

  public List<A> getWinners(int max) {
    return scores.getTop(max);
  }
}