/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.*;

import static java.util.Arrays.asList;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

public class Bayes implements Displayable, InspectableParent, ToTree, FromTree {
  private MultiSet<String> hamTable = new MultiSet<String>();
  private MultiSet<String> spamTable = new MultiSet<String>();
  private int hamCorpus, spamCorpus;
  private BayesParams params;

  public Bayes() {
    params = new BayesParams();
  }

  public Bayes(BayesParams params) {
    this.params = params;
  }

  public void addHamWord(String word, int count) {
    hamTable.add(word, count);
  }

  public void addSpamWord(String word, int count) {
    spamTable.add(word, count);
  }

  public void addHamCorpus(int count) {
    hamCorpus += count;
  }

  public void addSpamCorpus(int count) {
    spamCorpus += count;
  }

  public double getProbability(String word) {
    double g = hamTable.get(word), b = spamTable.get(word);
    /*
    g = Math.max(2, g);
    b = Math.max(2, b);
    double g2 = Math.min(1.0, g / hamCorpus);
    double b2 = Math.min(1.0, b / spamCorpus);
    return Math.max(0.01, Math.min(0.99, b2 / (g2 + b2)));
    */
    //if (g+b < 2) return 0.5;
    if (g+b == 0) return 0.5;
    double g2 = Math.min(1.0, g / hamCorpus);
    double b2 = Math.min(1.0, b / spamCorpus);
    return Math.max(0.01, Math.min(0.99, b2 / (g2 + b2)));
  }

  public double getCombinedProbability(Collection<String> words) {
    //int n = words.length;
    double a = 1.0, b = 1.0;
    /*double[] probabilities = new double[n];
    for (int i = 0; i < n; i++)
      probabilities[i] = getProbability(words[i]);*/
    for (String word : words) {
      double p = getProbability(word);
      if (Math.abs(p-0.5) >= params.getMinInterestingness()) {
        a *= p;
        b *= 1-p;
      }
    }
    return a/(a+b);
  }

  public int getHamCorpus() {
    return hamCorpus;
  }

  public int getSpamCorpus() {
    return spamCorpus;
  }

  public Set<String> getWords() {
    TreeSet<String> result = new TreeSet<String>();
    result.addAll(hamTable.asSet());
    result.addAll(spamTable.asSet());
    return result;
  }

  public int getHamCount(String word) {
    return hamTable.get(word);
  }

  public int getSpamCount(String word) {
    return spamTable.get(word);
  }

  public void display(Display display) {
    displayTable(display, "ham", hamCorpus, hamTable);
    displayTable(display, "spam", spamCorpus, spamTable);
  }

  private void displayTable(Display display, String name, int corpus, MultiSet<String> table) {
    display.beginSubsection(name + " (corpus size: " + corpus + ")");
    for (String word : table.getTopTen(10)) {
      display.put(word).put(" ").put(table.get(word)).nl();
    }
    display.endSubsection();
  }

  public BayesParams getParams() {
    return params;
  }

  public void simplify(int maxWords) {
    hamTable = simplify(hamTable, maxWords);
    spamTable = simplify(spamTable, maxWords);
  }

  private MultiSet<String> simplify(MultiSet<String> table, int maxWords) {
    MultiSet<String> result = new MultiSet<String>();
    for (String word : table.getTopTen(maxWords)) {
      result.add(word, table.get(word));
    }
    return result;
  }

  public String tableSizes() {
    return hamTable.asSet().size() + "/" + spamTable.asSet().size();
  }

  public Tree toTree() {
    return new Tree(getClass())
      .add("params", params.toTree())
      .add("hamCorpus", hamCorpus)
      .add("spamCorpus", spamCorpus)
      .add("hamTable", hamTable.toTree())
      .add("spamTable", spamTable.toTree());
  }

  public void fromTree(Tree tree) {
    params.fromTree(tree.get("params"));
    hamCorpus = tree.getInt("hamCorpus");
    spamCorpus = tree.getInt("spamCorpus");
    hamTable.fromTree(tree.get("hamTable"));
    spamTable.fromTree(tree.get("spamTable"));
  }

  public List<InspectionNode> getInspectionChildren() {
    return asList(new InspectionNode("hamTable", hamTable),
      new InspectionNode("spamTable", spamTable));
  }
}