/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.*;
import static drjava.util.TreeUtil.*;
import prophecy.common.Verbalizer;

import java.util.ArrayList;
import static java.util.Arrays.asList;
import java.util.Collection;
import java.util.List;
import java.util.TreeSet;

public class BayesSubsetContestant<A> implements SubsetContestant<A>, Displayable, InspectableParent, ToTree, FromTree {
  Bayes bayes;
  private int maxWords = 0;
  private BayesParams bayesParams;
  private List<Verbalizer<A>> verbalizers = new ArrayList<Verbalizer<A>>();

  public BayesSubsetContestant() {
    this(new Verbalizer[0]);
  }

  public BayesSubsetContestant(Verbalizer<A>... verbalizers) {
    this(new BayesParams(), asList(verbalizers));
  }

  public BayesSubsetContestant(List<Verbalizer<A>> verbalizers) {
    this(new BayesParams(), verbalizers);
  }

  public BayesSubsetContestant(BayesParams bayesParams, List<Verbalizer<A>> verbalizers) {
    this.bayesParams = bayesParams;
    this.verbalizers = verbalizers;
    bayes = new Bayes(bayesParams);
  }

  public void setVerbalizers(List<Verbalizer<A>> verbalizers) {
    this.verbalizers = verbalizers;
  }

  public boolean contains(A a) {
    Collection<String> words = getWords(a);
    double p = bayes.getCombinedProbability(words);
    return p < 0.5;
  }

  private Collection<String> getWords(A a) {
    TreeSet<String> words = new TreeSet<String>();
    for (Verbalizer<A> verbalizer : verbalizers) {
      words.addAll(verbalizer.getWords(a));
    }
    return words;
  }

  public void train(List<SubsetTI<A>> data) {
    trainVerbalizers(data);
    trainBayes(data);
  }

  private void trainVerbalizers(List<SubsetTI<A>> data) {
    List<A> examples = new ArrayList<A>();
    for (SubsetTI<A> item : data) {
      examples.add(item.example);
    }
    for (Verbalizer<A> verbalizer : verbalizers) {
      verbalizer.train(examples);
    }
  }

  private void trainBayes(List<SubsetTI<A>> data) {
    bayes = new Bayes(bayesParams);
    for (SubsetTI<A> item : data) {
      Collection<String> words = getWords(item.example);
      //System.out.println(item.inSet + " " + StringUtil.join(words, " "));
      if (item.inSet) {
        bayes.addHamCorpus(1);
        for (String word : words)
          bayes.addHamWord(word, 1);
      } else {
        bayes.addSpamCorpus(1);
        for (String word : words)
          bayes.addSpamWord(word, 1);
      }
    }

    if (maxWords != 0)
      bayes.simplify(maxWords);
  }

  public void display(Display display) {
    display.putnl(toString()).put(bayes);
  }

  public String toString() {
    String s = "BayesSC " + bayes.getParams();
    for (Verbalizer<A> verbalizer : verbalizers) {
      s += " " + ObjectUtil.toNiceString(verbalizer);
    }
    return s + " bayes: " + bayes.tableSizes();
  }

  public Tree toTree() {
    Tree tree = new Tree(getClass())
      .add("bayes", objectToTree(bayes))
      .add("verbalizers", listToTree(verbalizers));
    if (maxWords != 0)
      tree.add("maxWords", maxWords);
    return tree;
  }

  public void fromTree(Tree tree) {
    bayes = (Bayes) treeToObject(tree.get("bayes"));
    verbalizers = treeToList(tree.get("verbalizers"));
    maxWords = tree.getInt("maxWords", 0);
  }

  public List<InspectionNode> getInspectionChildren() {
    return asList(new InspectionNode(bayes));
  }
}