/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.*;

import static drjava.util.TreeUtil.objectToTree;
import static drjava.util.TreeUtil.treeToObject;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

public class ContestUtil {
  private static final String contestDir = "contests/";
  private static final String hallOfFameDir = "halloffame/";

  public static <A> void saveContest(IContest<A> contest, String name) {
    contestPersistence(name).store(objectToTree(contest));
  }

  private static FileTreePersistence contestPersistence(String name) {
    return new FileTreePersistence(contestDir + name + ".tree");
  }

  private static FileTreePersistence contestInfoPersistence(String name) {
    return new FileTreePersistence(contestDir + name + ".info.tree");
  }

  public static <A> Contest<A> loadContest(String name) {
    return treeToObject(contestPersistence(name).load());
  }

  public static <A> void saveResults(IContestRun<A> contest, String name) {
    String prefix = hallOfFameDir + name + "_";
    try {
      FileOutputStream fos = new FileOutputStream(prefix + "results.txt");
      contest.display(new StreamDisplay(new PrintStream(fos)));
      fos.close();
    } catch (Throwable e) {
      Errors.add(e);
    }

    try {
      if (contest.getWinner() != null)
        new FileTreePersistence(hallOfFameDir + name +"_winner.tree").store(objectToTree(contest.getWinner()));
    } catch (Throwable e) {
      Errors.add(e);
    }
  }

  public static <A> List<Subset<A>> runAndGetWinners(IContest<Subset<A>> contest, List<Subset<A>> contestants, int num) {
    ContestRun<Subset<A>> subContest = new ContestRun<Subset<A>>(contest);
    subContest.addContestants(contestants);
    subContest.run();
    return subContest.getWinners(num);
  }

  public static <A> Subset<A> runAndGetWinner(IContest<Subset<A>> contest, List<Subset<A>> contestants) {
    ContestRun<Subset<A>> subContest = new ContestRun<Subset<A>>(contest);
    subContest.addContestants(contestants);
    subContest.run();
    return subContest.getWinner();
  }

  public static Object loadWinner(String name) {
    Object winner = TreeUtil.treeToObject(new FileTreePersistence("halloffame/" + name + "_winner.tree").load());
    if (winner == null) {
      HallOfFame hallOfFame = (HallOfFame) hallOfFamePersistence(name).loadObject();
      if (hallOfFame != null)
        winner = hallOfFame.getWinner();
    }
    return winner;
  }

  public static <A> OpenContestRun<A> getOpenContestRun(String contestName) {
    Contest<A> contest = loadContest(contestName);
    if (contest == null)
      throw new RuntimeException("Contest " + contestName + " not found");
    return new OpenContestRun<A>(contestName, contest);
  }

  private static FileTreePersistence hallOfFamePersistence(String contestName) {
    return new FileTreePersistence(hallOfFameDir + "halloffame_" + contestName + ".tree");
  }

  public static <A> HallOfFame<A> loadHallOfFame(String contestName) {
    HallOfFame hallOfFame = TreeUtil.treeToObject(HallOfFame.class, hallOfFamePersistence(contestName).load());
    return hallOfFame != null ? hallOfFame : new HallOfFame<A>();
  }

  public static <A> void saveHallOfFame(String contestName, Scores<A> hallOfFame) {
    hallOfFamePersistence(contestName).store(objectToTree(hallOfFame));
  }

  public static HallOfFame<Tree> loadTreeHallOfFame(String contestName) {
    Tree tree = hallOfFamePersistence(contestName).load();
    return tree == null ? new HallOfFame<Tree>() : loadTreeHallOfFame(tree);
  }

  private static HallOfFame<Tree> loadTreeHallOfFame(Tree tree) {
    Tree scores = tree.get("scores");
    for (Tree t : scores.arguments()) {
      if (t.nameIs("#"))
        t.put(1, new Tree(Tree.class).add(t.get(1)));
    }
    return new HallOfFame<Tree>(tree);
  }

  public static <A> double getScore(Contest<A> contest, A contestant) {
    ContestRun<A> contestRun = new ContestRun<A>(contest);
    contestRun.addContestant(contestant);
    contestRun.runSilently();
    return contestRun.getScores().getScore(contestant);
  }

  public static <A> HallOfFame<A> loadHallOfFame(ContestOnDisk contestOnDisk) {
    return loadHallOfFame(contestOnDisk.getFullName());
  }

  public static List<ContestOnDisk> getAllContestsOnDisk() {
    return getAllContestsOnDisk("");
  }

  public static List<ContestOnDisk> getAllContestsOnDisk(String prefix) {
    List<ContestOnDisk> result = new ArrayList<ContestOnDisk>();
    for (File file : new File(contestDir).listFiles()) {
      if (file.isFile() && file.getName().startsWith(prefix)
        && file.getName().endsWith(".tree") && !file.getName().endsWith(".info.tree"))
        result.add(new ContestOnDisk(file.getName().substring(0, file.getName().length()-5)));
    }
    return result;
  }

  public static ContestInfo loadContestInfo(String name) {
    ContestInfo contestInfo = (ContestInfo) contestInfoPersistence(name).loadObject();
    return contestInfo == null ? new ContestInfo() : contestInfo;
  }

  public static void saveContestInfo(ContestInfo contestInfo, String name) {
    contestInfoPersistence(name).store(contestInfo);
  }
}