/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.SortedList;
import drjava.util.Tree;
import prophecy.common.Scores;

import static java.lang.Math.*;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

public class HallOfFame<A> extends Scores<A> {
  private Integer maxSize = 10;
  private Double minDiff = 0.01;

  private SortedList<A> sortedList = new SortedList<A>(new Comparator<A>() {
    public int compare(A c1, A c2) {
      return (int) signum(getScore(c2)-getScore(c1));
    }
  });

  public HallOfFame() {
  }

  public HallOfFame(Scores<A> scores) {
    for (A a : scores.getContestants()) {
      add(a, scores.getScore(a));
    }
  }

  public void add(A a, double score) {
    super.add(a, score); // so comparator works correctly for the next few lines
    boolean keep = false;
    try {
      if (minDiff != null) {
        int i = sortedList.getInsertionIndex(a);
        if (i > 0) {
          double diff = abs(getScore(sortedList.get(i - 1)) - score);
          if (diff < minDiff)
            return;
        }
        if (i < sortedList.size()) {
          double s = getScore(sortedList.get(i));
          double diff = abs(s - score);
          if (score < s)
            throw new RuntimeException("Wrong order");
          if (diff < minDiff)
            return;
        }
      }
      keep = true;
    } finally {
      if (!keep)
        removeContestant(a);
      else {
        sortedList.add(a);
        if (maxSize != null && sortedList.size() > maxSize)
          removeContestant(sortedList.last());
        check();
      }
    }
  }

  private void check() {
    if (sortedList.size() != getContestants().size())
      throw new RuntimeException("Size mismatch: " + sortedList.size() + " vs " + getContestants().size());
  }

  public void removeContestant(A a) {
    super.removeContestant(a);
    sortedList.remove(a);
  }

  public HallOfFame(Tree tree) {
    this(new Scores<A>(tree));
  }

  public List<A> getTop(int max) {
    return sortedList.subList(0, min(sortedList.size(), max));
  }

  public void setMinDiff(Double minDiff) {
    this.minDiff = minDiff;
  }

  public int size() {
    return sortedList.size();
  }

  public List<A> getList() {
    return Collections.unmodifiableList(sortedList);
  }
}