/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.Function;
import static drjava.util.MultiCoreUtil.getNumCores;
import static drjava.util.MultiCoreUtil.parallelMap;
import drjava.util.ObjectUtil;
import drjava.util.Tree;
import drjava.util.TreeUtil;
import prophecy.common.SingletonGenerator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class OpenContestRun<A> implements ContestantReceiver<A>, InspectableParent {
  private Contest<A> contest;
  private String contestName;
  HallOfFame<A> hallOfFame;
  List<Generator<? extends A>> generators = new ArrayList<Generator<? extends A>>();
  private OpenContestUI<A> ui = new OpenContestSilentUI<A>();
  private boolean reevaluate;
  //private boolean generatorInitialized;
  //private List<A> contestants = new ArrayList<A>();

  public OpenContestRun(String contestName, Contest<A> contest) {
    this.contest = contest;
    this.contestName = contestName;
    hallOfFame = ContestUtil.loadHallOfFame(contestName);
    if (hallOfFame == null) hallOfFame = new HallOfFame<A>();
  }

  public OpenContestRun(String contestName) {
    this(contestName, ContestUtil.<A> loadContest(contestName));
  }

  public OpenContestRun(ContestOnDisk contestOnDisk) {
    this(contestOnDisk.getFullName());
  }

  public void setUI(OpenContestUI<A> ui) {
    this.ui = ui;
    ui.setContestName(contestName);
    ui.showHallOfFame(hallOfFame);
  }

  public void run() {
    while (true) {
      List<A> contestants = new ArrayList<A>();
      for (int i = 0; i < getNumCores(); i++) {
        A contestant = produceContestant();
        if (contestant != null) {
          contestants.add(contestant);
          ui.newContestant(i, ObjectUtil.toNiceString(contestant));
        } else
          break;
      }

      if (contestants.isEmpty())
        break;

      List<Double> scores = parallelMap(contestants, new Function<A, Double>() {
        public Double get(A contestant) {
          //contest.trainContestant(contestant);
          return evaluateContestant(contestant);
        }
      });

      for (int i = 0; i < contestants.size(); i++) {
        hallOfFame.add(contestants.get(i), scores.get(i));
      }
      ui.showHallOfFame(hallOfFame);
      ContestUtil.saveHallOfFame(contestName, hallOfFame);
    }

    ui.done();
  }

  private A produceContestant() {
    A contestant;
    do {
      contestant = produceContestant2();
      if (contestant == null) return null;
    } while (!reevaluate && alreadyProduced(contestant));
    return contestant;
  }

  private boolean alreadyProduced(A contestant) {
    Tree tree = TreeUtil.objectToTree(contestant);
    for (A c : hallOfFame.getContestants()) {
      if (TreeUtil.equal(TreeUtil.objectToTree(c), tree))
        return true;
    }
    return false;
  }

  private A produceContestant2() {
    while (true) {
      if (generators.isEmpty()) return null;
      /*if (!generatorInitialized) {
        generators.get(0).init();
        generatorInitialized = true;
      }*/
      A a = generators.get(0).next();
      if (a != null) return a;
      generators.remove(0);
      //generatorInitialized = false;
    }
  }

  private double evaluateContestant(A contestant) {
    ContestRun<A> contestRun = new ContestRun<A>(contest);
    contestRun.addContestant(contestant);
    contestRun.run();

    // this is not threadsafe, but what the heck
    hallOfFame.setBestPossibleScore(contestRun.getScores().getBestPossibleScore());
    hallOfFame.setWorstPossibleScore(contestRun.getScores().getWorstPossibleScore());

    return contestRun.getScores().getScore(contestant);
  }

  public void addGenerator(Generator<? extends A> generator) {
    generators.add(generator);
  }

  public Contest<A> getContest() {
    return contest;
  }

  public void addContestant(A a) {
    addGenerator(new SingletonGenerator<A>(a));
  }

  public void clearHallOfFame() {
    hallOfFame = new HallOfFame<A>();
  }

  public void runWithStandardUI() {
    setUI(new OpenContestRetroUI<A>());
    run();
  }

  public List<InspectionNode> getInspectionChildren() {
    return Arrays.asList(
      new InspectionNode("contest", contest),
      new InspectionNode("hallOfFame", hallOfFame));
  }

  public String toString() {
    return contestName;
  }

  public boolean isReevaluate() {
    return reevaluate;
  }

  public void setReevaluate(boolean reevaluate) {
    this.reevaluate = reevaluate;
  }
}