/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.vats.FulfilledPromise;
import drjava.vats.Promise;
import drjava.vats.PromiseListener;
import sorttreetable.SortTableComparator;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import java.util.*;

public abstract class PromiseBasedTableModel<A> extends AbstractTableModel implements TableModelListener {
  List<A> list = new ArrayList<A>();
  protected int sortColumn = -1;
  protected boolean sortAscending;
  protected Promise<Object>[][] cellData;
  protected String[] columnNames;
  private boolean settingCellData;

  protected PromiseBasedTableModel(String[] columnNames) {
    this.columnNames = columnNames;
    addTableModelListener(this);
  }

  protected PromiseBasedTableModel(String[] columnNames, Collection<A> collection) {
    this(columnNames);
    list = new ArrayList<A>(collection);
  }

  protected abstract int compareRows(A a, A b);

  protected String getCellQuick(A a, int row, int column) {
    Object o = getCell(a, row, column);
    if (o instanceof String) return (String) o;
    if (o instanceof Promise && ((Promise<String>) o).isFulfilled())
      return ((Promise<String>) o).getValue();
    return "";
  }

  public void setList(Collection<A> collection) {
    list = new ArrayList<A>(collection);
    sort();
    fireTableDataChanged();
  }

  public int getRowCount() {
    return list.size();
  }

  public Object getValueAt(final int row, final int col) {
    Promise<Object> promise = getCellData(row, col);
    if (promise == null) {
      Object o = getCell(list.get(row), row, col);
      if (o instanceof Promise)
        promise = (Promise<Object>) o;
      else
        promise = new FulfilledPromise<Object>(o);
      cellData[row][col] = promise;
      if (!promise.isFulfilled())
        promise.whenResolved(new PromiseListener<Object>() {
          public void fulfilled(Object o) {
            resolved();
          }

          public void broken(Throwable throwable) {
            resolved();
          }

          private void resolved() {
            //System.out.println(row + ":" + col + " resolved in turn " + Vat.currentVat().getTurns());
            settingCellData = true;
            try {
              fireTableCellUpdated(row, col);
            } finally {
              settingCellData = false;
            }
          }
        });
    }

    if (promise.isFulfilled())
      return promise.getValue();
    else
      return "";
  }

  private Promise<Object> getCellData(int row, int col) {
    if (cellData == null) cellData = new Promise[getRowCount()][getColumnCount()];
    return cellData[row][col];
  }

  /** may return: Promise<String>, String, Icon */
  protected abstract Object getCell(A a, int row, int col);

  public List<A> getList() {
    return list;
  }

  public void sortColumn(final int col, final boolean ascending) {
    sortColumn = col;
    sortAscending = ascending;
    sort();
  }

  protected void sort() {
    if (sortColumn == -1) return;
    Collections.sort(list, new Comparator<A>() {
      public int compare(A a, A b) {
        int result = compareRows(a, b);
        return sortAscending ? result : -result;
      }
    });
    fireTableDataChanged();
  }

  public SortTableComparator getComparator() {
    throw new UnsupportedOperationException();
  }

  public void setComparator(SortTableComparator sorter) {
    throw new UnsupportedOperationException();
  }

  public int getColumnCount() {
    return columnNames.length;
  }

  protected void setColumnNames(String[] columnNames) {
    this.columnNames = columnNames;
  }

  public String getColumnName(int i) {
    return columnNames[i];
  }

  public void tableChanged(TableModelEvent e) {
    if (!settingCellData) {
      //System.out.println("event type: " + e.getType());
      //if (e.getType() != TableModelEvent.UPDATE)
        cellData = new Promise[getRowCount()][getColumnCount()];
    }
  }

  public A getItem(int row) {
    return row >= 0 && row < list.size() ? list.get(row) : null;
  }

  public void addItem(A a) {
    list.add(a);
    fireTableRowsInserted(list.size()-1, list.size()-1);
  }

  public void removeItem(int row) {
    list.remove(row);
    fireTableRowsDeleted(row, row);
  }
}