/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common;

import drjava.util.*;
import prophecy.common.InspectableParent;
import prophecy.common.InspectionNode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class Scores<A> implements ToTree, Displayable, InspectableParent {
  private AnalogSet<A> scores;
  private Double worstPossibleScore, bestPossibleScore;

  public Scores() {
    scores = new AnalogSet<A>();
    scores.setAllowZero(true);
  }

  public Double getWorstPossibleScore() {
    return worstPossibleScore;
  }

  public void setWorstPossibleScore(Double worstPossibleScore) {
    this.worstPossibleScore = worstPossibleScore;
  }

  public Double getBestPossibleScore() {
    return bestPossibleScore;
  }

  public void setBestPossibleScore(Double bestPossibleScore) {
    this.bestPossibleScore = bestPossibleScore;
  }

  public void add(A a, double score) {
    scores.add(a, score);
  }

  public void display(Display display) {
    List<A> list = scores.sortedByDescValue();
    if (list.isEmpty())
      display.putnl("(No contestants)");
    for (int i = 0; i < list.size(); i++) {
      A contestant = list.get(i);
      String s = formatScoreLine(contestant);
      display.putnl("#" + (i + 1) + s);
    }
  }

  public String formatScoreLine(A contestant) {
    String s = formatScore(contestant);
    return s + ObjectUtil.toNiceString(contestant);
  }

  public String formatScore(A contestant) {
    double score = scores.get(contestant);
    String scoreText;
    if (worstPossibleScore != null && bestPossibleScore != null && worstPossibleScore < bestPossibleScore)
      scoreText = StringUtil.formatProbability((score-worstPossibleScore)/(bestPossibleScore-worstPossibleScore), 1) + "%";
    else
      scoreText = StringUtil.formatDouble(score, 2);
    String s = (bestPossibleScore != null && score >= bestPossibleScore ? " PERFECT" : "")
      + " [" + scoreText + "] ";
    return s;
  }

  public A getWinner() {
    return scores.isEmpty() ? null : scores.sortedByDescValue().get(0);
  }

  public double getScore(A a) {
    return scores.get(a);
  }

  public boolean contains(A a) {
    return scores.contains(a);
  }

  public List<A> getTop(int max) {
    List<A> list = scores.sortedByDescValue();
    return list.size() > max ? list.subList(0, max) : list;
  }

  public Tree toTree() {
    Tree tree = new Tree(this.getClass()).add("scores", scores.toTree());
    tree.add("worstPossibleScore", worstPossibleScore);
    tree.add("bestPossibleScore", bestPossibleScore);
    return tree;
  }

  public Scores(Tree tree) {
    this();
    scores.fromTree(tree.get("scores"));
    worstPossibleScore = tree.getDouble("worstPossibleScore");
    bestPossibleScore = tree.getDouble("bestPossibleScore");
  }

  public Collection<A> getContestants() {
    return scores.asSet();
  }

  public void removeContestant(A a) {
    scores.remove(a);
  }

  public List<InspectionNode> getInspectionChildren() {
    List<InspectionNode> result = new ArrayList<InspectionNode>();
    for (A a : scores.asSet()) {
      result.add(new InspectionNode(formatScoreLine(a), a));
    }
    return result;
  }
}