/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common.lifeform;

import drjava.util.*;
import static drjava.util.ObjectUtil.nice;
import static drjava.util.StringUtil.n;
import prophecy.common.Prophecy;
import drjava.util.Errors;
import prophecy.common.Flag;
import prophecy.common.Visualizable;
import prophecy.m.m14.Failure;
import prophecy.m.m20.ReleaseListener;

import javax.swing.*;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

public class Lifeform implements Visualizable {
  private static long objectCounter;

  private final long objectID;
  private Date birthdate = new Date();
  protected Logger logger;
  private String status = "new";
  private Environment environment = null;
  private static Randomizer randomizer = new RealRandomizer();
  private Flag retirementFlag = new Flag();
  private double cpuTime;
  protected Activity activity;
  private List<LogEntry> log;
  private boolean isLeader;
  private Extensions extensions = new Extensions();
  private List<Object> attachments = new ArrayList<Object>();
  private boolean ranOnce;

  public Lifeform() {
    objectID = newID();
    log = new ArrayList<LogEntry>();
    logger = new MemoryLogger(log);
    initSmartIntroduce();

    logger.log("Object created");
    Prophecy.getGlobalCommunityServices().getHomelessShelter().add(this);
  }

  private void initSmartIntroduce() {
    SmartIntroduce smartIntroduce = new SmartIntroduce(getClass());
    if (smartIntroduce.getMethodCount() != 0)
      extensions.put(SmartIntroduce.class, smartIntroduce);
  }

  public void setLogger(Logger logger) {
    this.logger = logger;
  }

  private static synchronized long newID() {
    return ++objectCounter;
  }

  public ObjectID getObjectID() {
    return new ObjectID(objectID);
  }

  public Logger getLogger() {
    return logger;
  }

  /** override me */
  public String getStatus() {
    return status;
  }

  public void setStatus(String status) {
    if (!ObjectUtil.equal(status, this.status)) {
      this.status = status;
      log("Status: " + status);
    }
  }

  public void logSurprise(Object o) {
    logger.logSurprise(nice(o));
  }

  public void logActivity(String text) {
    logger.logActivity(text);
  }

  public void log(String text) {
    logger.log(text);
  }

  public void logConfirmation(String text) {
    logger.logConfirmation(text);
  }

  public void invokeLater(Runnable runnable) {
    new Thread(runnable).start();
  }

  /** override me */
  protected void idleImpl() {

  }

  public final void idle() {
    if (!ranOnce) {
      ranOnce = true;
      try {
        runOnce();
      } catch (Throwable e) {
        error(e);
      }
    }

    if (activity != null)
      try {
        activity.idle();
      } catch (Throwable e) {
        Errors.add(e);
      }
    idleImpl();
  }

  /** override me */
  protected void runOnce() {
  }

  /** override me */
  public void introduce(Lifeform lifeform) {
    if (activity != null)
      activity.introduce(lifeform);
    SmartIntroduce smartIntroduce = extensions.get(SmartIntroduce.class);
    if (smartIntroduce != null)
      smartIntroduce.introduce(this, lifeform);
  }

  public Environment getEnvironment() {
    return environment;
  }

  public void setEnvironment(Environment environment) {
    if (this.environment == null && environment != null)
      Prophecy.getGlobalCommunityServices().getHomelessShelter().remove(this);
    else if (this.environment != null && environment == null)
      Prophecy.getGlobalCommunityServices().getHomelessShelter().add(this);
    this.environment = environment;
  }

  public static Randomizer getRandomizer() {
    return randomizer;
  }

  public void release(Lifeform lifeform) {
    if (environment == null)
      logSurprise("Can't release " + lifeform + ", no environment");
    else {
      logActivity("Releasing " + lifeform);
      environment.addInhabitant(lifeform);
      try {
        introduce(lifeform);
      } catch (Throwable e) {
        Errors.add(e);
      }
      for (ReleaseListener listener : getReleaseListeners())
        try {
          listener.lifeformReleased(new ReleaseEvent(this, lifeform));
        } catch (Throwable e) {
          Errors.add(e);
        }
    }
  }

  public String getAgeAsText() {
    long ms = new Date().getTime() - getBirthdate().getTime();
    int minutes = (int) (ms/1000/60);
    return n(minutes, "minute");
  }

  public Date getBirthdate() {
    return birthdate;
  }

  public void retire() {
    retirementFlag.raise();
  }

  public Flag getRetirementFlag() {
    return retirementFlag;
  }

  public String toString() {
    return /*"#" + objectID + " " +*/ ObjectUtil.niceify(this, super.toString());
  }

  public void addCPUTime(long ms) {
    cpuTime += ms*0.001;
  }

  public double getCpuTime() {
    return cpuTime;
  }

  public JComponent visualize() {
    return null;
  }

  public void error(Throwable throwable) {
    setStatus("Error: " + throwable);
    Errors.add(throwable);
  }

  public void error(String text) {
    setStatus("Error: " + text);
  }

  public void setActivity(Activity activity) {
    if (this.activity != null)
      this.activity.setLifeform(null);
    this.activity = activity;
    if (activity != null) {
      setStatus(activity.getStatus());
      activity.setLifeform(this);
    }
  }

  public Activity getActivity() {
    return activity;
  }

  public List<LogEntry> getLog() {
    return log;
  }

  public boolean isLeader() {
    return isLeader;
  }

  public void setLeader(boolean leader) {
    this.isLeader = leader;
    if (leader)
      log("Now leader");
    else
      log("No longer leader");
  }

  /** overridable */
  public boolean wantsToLead() {
    return false;
  }

  public Icon getIcon() {
    if (extensions.get(Icon.class) == null)
      try {
        setIcon(makeIcon());
      } catch (Throwable e) {
        error(e);
      }
    return extensions.get(Icon.class);
  }

  /** overridable */
  public Icon makeIcon() {
    return null;
  }

  public void setIcon(Icon icon) {
    extensions.put(Icon.class, icon);
  }

  public boolean isSomebody() {
    return true;
  }

  public List<Lifeform> getPeers() {
    return environment == null ? new ArrayList<Lifeform>() : environment.getInhabitants();
  }

  public <A> List<A> getPeers(Class<A> peerClass) {
    List<A> result = new ArrayList<A>();
    for (Lifeform lifeform : getPeers())
      if (peerClass.isInstance(lifeform))
        result.add((A) lifeform);
    return result;
  }

  /** overridable */
  public List<String> getDefiningProperties() {
    return Lizt.of();
  }

  /** overridable */
  public List<String> getCoreFunctions() {
    return Lizt.of();
  }

  public boolean isRetired() {
    return retirementFlag.isUp();
  }

  public boolean hasAttachment(Class attachmentClass) {
    for (Object attachment : attachments) {
      if (attachmentClass.isInstance(attachment))
        return true;
    }
    return false;
  }

  public void addAttachment(Object attachment) {
    attachments.add(attachment);
  }

  public String getShortenedClassName() {
    return ObjectUtil.shortenClassName(getClass().getName());
  }

  public Time getCurrentTime() {
    TimeService timeService = getTimeService();
    return timeService == null ? null : timeService.getCurrentTime();
  }

  public TimeService getTimeService() {
    return environment == null ? null : environment.getTimeService();
  }

  public void addReleaseListener(ReleaseListener listener) {
    ReleaseListeners listeners = extensions.get(ReleaseListeners.class);
    if (listeners == null)
      extensions.put(ReleaseListeners.class, listeners = new ReleaseListeners());
    listeners.add(listener);
  }

  private List<ReleaseListener> getReleaseListeners() {
    ReleaseListeners listeners = extensions.get(ReleaseListeners.class);
    return listeners == null ? new ArrayList<ReleaseListener>() : listeners;
  }

  public void fail(Throwable throwable) {
    logSurprise(throwable);
    throw new Failure(nice(throwable));
  }

  public void fail(String msg) {
    logSurprise(msg);
    throw new Failure(msg);
  }
}