/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common.socket;

import drjava.util.Errors;
import drjava.util.Log;
import prophecy.common.socket.ObjectURL;
import prophecy.t.t03.HasURL;
import prophecy.t.t11.SocketGuard;

import java.io.IOException;
import java.net.BindException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;

// from: t03
public class DialogServer implements HasURL {
  private int port;
  private ServerSocket serverSocket;
  private SocketCallable callHandler;
  private Thread acceptThread;
  private boolean accepting;
  private List<SocketHandler> socketHandlers = Collections.synchronizedList(new ArrayList<SocketHandler>());

  public DialogServer(int port, SocketCallable callHandler) {
    this.port = port;
    this.callHandler = callHandler;
  }

  public DialogServer(ServerSocket serverSocket, SocketCallable callHandler) {
    this.serverSocket = serverSocket;
    port = serverSocket.getLocalPort();
    this.callHandler = callHandler;
  }

  public void start() throws IOException {
    if (serverSocket == null)
      try {
        serverSocket = new ServerSocket(port);
      } catch (BindException e) {
        if ("Address already in use: JVM_Bind".equals(e.getMessage()))
          throw new IOException("Port " + port + " occupied (server started already)?");
        else
          throw e;
      }

    Log.info("DialogServer listening: " + getURL());
    acceptConcurrently();
  }

  public ObjectURL getURL() {
    return new ObjectURL("obj:socket[" + port + "]");
  }

  private void acceptConcurrently() {
    accepting = true;
    acceptThread = new Thread("DialogServer accept") {
      public void run() {
        try {
          while (true) {
            final Socket socket = serverSocket.accept();
            handleConcurrently(socket);
          }
        } catch (Throwable e) {
          if (accepting)
            Errors.add(e);
        }
      }
    };
    acceptThread.start();
  }

  private void handleConcurrently(final Socket socket) {
    new Thread("DialogServer client handler") {
      public void run() {
        try {
          handleClient(socket);
        } catch (IOException e) {
          Errors.add(e);
        }
      }
    }.start();
  }

  private void handleClient(Socket socket) throws IOException {
    final SocketHandler handler = new SocketHandler(socket);
    socketHandlers.add(handler);
    try {
      callHandler.takeCall(handler);
    } finally {
      handler.softClose();
      socketHandlers.remove(handler);
    }
  }

  public SocketCallable getCallHandler() {
    return callHandler;
  }

  public void stop() throws IOException {
    accepting = false;
    serverSocket.close(); // interrupts acceptThread
    /*if (acceptThread != null)
      acceptThread.interrupt();*/
  }

  public void setCallHandler(SocketCallable callHandler) {
    this.callHandler = callHandler;
  }

  public void addGuard(SocketGuard guard) {
    guard.setFollowUp(getCallHandler());
    setCallHandler(guard);
  }

  public void closeAllConnections() {
    while (!socketHandlers.isEmpty()) {
      SocketHandler socketHandler = socketHandlers.get(0);
      socketHandlers.remove(socketHandler);
      socketHandler.close();
    }
  }
}