/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common.socket;

import drjava.util.Bus;
import drjava.util.Errors;
import drjava.util.Tree;
import static drjava.util.ObjectUtil.nice;

import prophecy.t.t08.InOut;
import prophecy.t.t08.Packet;

import java.io.*;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;

public class SocketHandler {
  private InputStream inputStream;
  private BufferedReader bufferedReader;
  private OutputStream outputStream;
  private PrintWriter printWriter;
  private boolean lineSent;
  private boolean closed;
  private boolean dumpAll = true;
  private ISocket socket;
  private boolean keepOpen;
  private int charactersSent;
  private Bus<Packet> packetBus = new Bus<Packet>();
  private IOException ioException;

  public SocketHandler(Socket socket) throws IOException {
    this.socket = new RealSocket(socket);
    init(socket.getInputStream(), socket.getOutputStream());
  }

  private void init(InputStream inputStream, OutputStream outputStream) {
    this.inputStream = inputStream;
    this.outputStream = outputStream;
    bufferedReader = new BufferedReader(new InputStreamReader(inputStream));
    printWriter = new PrintWriter(outputStream, true);
  }

  public SocketHandler(ISocket socket) throws IOException {
    this.socket = socket;
    init(socket.getInputStream(), socket.getOutputStream());
  }

  public String readLine() {
    try {
      String line = bufferedReader.readLine();
      if (line != null) {
        if (dumpAll)
          System.out.println("< " + line);
        packetBus.receive(new Packet(InOut.in, line + "\n"));
      }
      return line;
    } catch (IOException e) {
      if (socket.isClosed())
        return null;
      ioException = e;
      throw new RuntimeException(e);
    }
  }

  public void println(String s) {
    printWriter.println(s);
    charactersSent += s.length()+1;
    lineSent = true;
    packetBus.receive(new Packet(InOut.out, s + "\n"));
    if (dumpAll)
      System.out.println("> " + s);
  }

  public void close() {
    if (!closed) {
      closed = true;
      try {
        socket.close();
        inputStream.close();
        outputStream.close();
      } catch (IOException e) {
        Errors.add(e);
      }
    }
  }

  public boolean lineSent() {
    return lineSent;
  }

  public String getInfo() {
    try {
      SocketAddress address = socket.getRemoteSocketAddress();
      return nice(address);
    } catch (Throwable exception) {
      Errors.add(exception);
      return "";
    }
  }

  public void softClose() throws IOException {
    if (!keepOpen)
      close();
  }

  public boolean isKeepOpen() {
    return keepOpen;
  }

  public void setKeepOpen(boolean keepOpen) {
    this.keepOpen = keepOpen;
  }

  public boolean isClosed() {
    return closed;
  }

  public int getCharactersSent() {
    return charactersSent;
  }

  public InetSocketAddress getRemoteAddress() {
    return (InetSocketAddress) socket.getRemoteSocketAddress();
  }

  public Bus<Packet> getPacketBus() {
    return packetBus;
  }

  // Utility methods

  public void println(Tree tree) {
    println(tree.toString());
  }

  public void ok(Tree tree) {
    println(new Tree("OK.").add(tree));
  }

  public void error(Tree tree) {
    println(new Tree("Error.").add(tree));
  }

  public Tree expectOK() {
    String line = readLine();
    if (!line.startsWith("OK"))
      throw new RuntimeException("OK expected: " + line);
    if (line.startsWith("OK.")) line = line.substring(3);
    else line = line.substring(2);
    line = line.trim();
    return line.length() == 0 ? null : Tree.parse(line);
  }

  public void request(Tree tree) {
    println(new Tree("Request.").add(tree));
  }

  public Tree readTree() {
    return Tree.parse(readLine());
  }

  public void ok() {
    println(new Tree("OK."));
  }

  public IOException getIOException() {
    return ioException;
  }
}