/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common.sound;

import drjava.util.PopupMenuHelper;
import prophecy.audio.a01.SoundFunction;
import prophecy.common.Marker;
import prophecy.common.Surface;
import static prophecy.common.image.PixelUtil.blend;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

public class SoundSurface extends Surface {
  private Sound sound;
  private List<Marker> markers = new ArrayList<Marker>();
  private double amplification = 1.0;

  /** function values are assumed to be between 0 and 1 */
  private List<SoundFunction> functions = new ArrayList<SoundFunction>();

  public SoundSurface() {
    setBackground(Color.black);

    new PopupMenuHelper() {
      protected void fillMenu(JPopupMenu menu, Point point) {
        JMenuItem miPlay = new JMenuItem("Play");
        miPlay.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            SoundUtil.playInExtraThread(sound);
          }
        });
        menu.add(miPlay);
      }
    }.install(this);
  }

  public SoundSurface(Sound sound) {
    this();
    this.sound = sound;
  }

  public SoundSurface(SoundWithMarkers soundWithMarkers) {
    this(soundWithMarkers.sound, soundWithMarkers.markers);
  }

  public SoundSurface(Sound sound, List<Marker> markers) {
    this.sound = sound;
    for (Marker marker : markers) {
      addMarker(marker);
    }
  }

  protected int secondToX(double second) {
    return (int) (second*getWidth()/sound.getSeconds());
  }

  private double xToSecond(int x) {
    return x*sound.getSeconds()/getWidth();
  }

  private int sampleToY(double sample) {
    return (int) ((1-sample*amplification)*0.5*getHeight());
  }

  public void render(int w, int h, Graphics2D g) {
    if (sound != null) {
      drawMarkers(g);
      drawSample(g);
      drawFunctions(g);
    }
  }

  private void drawFunctions(Graphics2D g) {
    int y = 0, xStep = 2;
    for (SoundFunction function : functions) {
      for (int x = 0; x < getWidth(); x += xStep) {
        double value = function.getValue(sound, xToSecond(x));
        g.setColor(blend(Color.black, Color.white, value).getColor());
        g.fillRect(x, y, xStep, 10);
      }
      y += 10;
    }
  }

  private void drawSample(Graphics2D g) {
    g.setColor(Color.yellow);
    for (int x = 0; x < getWidth(); x++) {
      Sound clip = sound.clip(new Marker(xToSecond(x)), new Marker(xToSecond(x + 1)));
      double min = getMin(clip), max = getMax(clip);
      if (max >= min)
        g.drawLine(x, sampleToY(min), x, sampleToY(max));
    }
  }

  private void drawMarkers(Graphics2D g) {
    g.setColor(Color.blue);
    FontMetrics fontMetrics = g.getFontMetrics();
    List<Point> points = new ArrayList<Point>();
    for (Marker marker : markers) {
      int x = secondToX(marker.getSecond());
      g.drawLine(x, 0, x, getHeight());
      if (marker.getName() != null && marker.getName().length() > 0) {
        int fontX = x+2, fontY = findY(points, fontX) + fontMetrics.getAscent()+2;
        g.drawString(marker.getName(), fontX, fontY);
        points.add(new Point(
          fontX + fontMetrics.stringWidth(marker.getName()) + 2, fontY + fontMetrics.getDescent()));
      }
    }
  }

  private int findY(List<Point> points, int x) {
    int y = 0;
    for (Point point : points) {
      if (point.x >= x)
        y = point.y;
    }
    return y;
  }

  private double getMin(Sound clip) {
    double min = 1.0;
    for (float sample : clip.getSamples()) {
      if (sample < min) min = sample;
    }
    return min;
  }

  private double getMax(Sound clip) {
    double max = -1.0;
    for (float sample : clip.getSamples()) {
      if (sample > max) max = sample;
    }
    return max;
  }

  public void addMarker(Marker marker) {
    markers.add(marker);
  }

  public void addFunction(SoundFunction function) {
    functions.add(function);
  }

  public void addMarkers(List<Marker> markers) {
    this.markers.addAll(markers);
  }

  public void setSound(Sound sound) {
    this.sound = sound;
    repaint();
  }

  public Sound getSound() {
    return sound;
  }

  public double getAmplification() {
    return amplification;
  }

  public void setAmplification(double amplification) {
    this.amplification = amplification;
    repaint();
  }
}