/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.common.sound;

import org.tritonus.share.sampled.FloatSampleBuffer;
import drjava.util.Errors;
import prophecy.common.Marker;

import javax.sound.sampled.*;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;

public class SoundUtil {
  public static byte[] readFully(AudioInputStream inputStream) throws IOException {
    int samples = (int) inputStream.getFrameLength();
    byte[] buffer = new byte[samples * inputStream.getFormat().getFrameSize()];
    int idx = 0, bytes;
    while ((bytes = inputStream.read(buffer, idx, buffer.length - idx)) > 0) {
      idx += bytes;
    }
    return buffer;
  }

  public static FloatSampleBuffer loadFloatSampleBuffer(File file) {
    try {
      AudioInputStream inputStream = AudioSystem.getAudioInputStream(file);
      byte[] tempBuffer = readFully(inputStream);
      inputStream.close();
      return new FloatSampleBuffer(tempBuffer, 0, tempBuffer.length, inputStream.getFormat());
    } catch (IOException e) {
      throw new RuntimeException(e);
    } catch (UnsupportedAudioFileException e) {
      throw new RuntimeException(e);
    }
  }

  public static Sound load(String file) {
    return load(new File(file));
  }

  public static Sound load(File file) {
    FloatSampleBuffer sampleBuffer = loadFloatSampleBuffer(file);
    while (sampleBuffer.getChannelCount() > 1)
      sampleBuffer.removeChannel(0, false);
    float[] audioData = sampleBuffer.getChannel(0);
    Sound sound = new Sound(sampleBuffer.getSampleRate(), audioData);
    sound.setFile(file.getPath());
    return sound;
  }

  public static List<Marker> makeMarkers(double[] seconds) {
    List<Marker> markers = new ArrayList<Marker>();
    for (double second : seconds) {
      markers.add(new Marker(second));
    }
    return markers;
  }

  public static Sound loadSound(String file) {
    return load(new File(file));
  }

  public static void save(Sound sound, String fileName) {
    AudioFormat format = stdAudioFormat(sound);
    byte[] bytes = getBytes(sound, format);
    try {
      AudioSystem.write(new AudioInputStream(new ByteArrayInputStream(bytes), format,
        sound.getNumSamples()), AudioFileFormat.Type.WAVE, new File(fileName));
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  private static byte[] getBytes(Sound sound, AudioFormat format) {
    FloatSampleBuffer buffer = new FloatSampleBuffer(1, sound.getNumSamples(), (float) sound.getSampleRate());
    System.arraycopy(sound.getSamples(), 0, buffer.getChannel(0), 0, sound.getNumSamples());
    byte[] bytes = buffer.convertToByteArray(format);
    return bytes;
  }

  private static AudioFormat stdAudioFormat(Sound sound) {
    AudioFormat format = new AudioFormat((float) sound.getSampleRate(), 16, 1, true, false);
    return format;
  }

  public static void play(Sound sound) {
    try {
      AudioFormat format = stdAudioFormat(sound);
      byte audio[] = getBytes(sound, format);

      InputStream input = new ByteArrayInputStream(audio);
      final AudioInputStream ais = new AudioInputStream(input, format,
        audio.length / format.getFrameSize());
      int desiredBufferSize = (int) (0.5 * format.getFrameRate() * format.getFrameSize());
      SourceDataLine.Info info = new DataLine.Info(SourceDataLine.class, format,
        desiredBufferSize);
      final SourceDataLine line = (SourceDataLine) AudioSystem.getLine(info);
      line.open(format);
      line.start();

      int offset = 0;
      while (offset < audio.length) {
        offset += line.write(audio, offset, audio.length-offset);
      }
      line.drain();
      line.close();
      input.close();
    }
    catch (Exception e) {
      Errors.add(e);
    }
  }

  public static void playInExtraThread(final Sound sound) {
    new Thread() {
      public void run() {
        SoundUtil.play(sound);
      }
    }.start();
  }

  public static Sound silence(double sampleRate, double seconds) {
    return new Sound(sampleRate, seconds);
  }

  public static Sound append(Sound a, Sound b) {
    float[] samples = new float[a.getNumSamples() + b.getNumSamples()];
    System.arraycopy(a.getSamples(), 0, samples, 0, a.getNumSamples());
    System.arraycopy(b.getSamples(), 0, samples, a.getNumSamples(), b.getNumSamples());
    return new Sound(a.getSampleRate(), samples);
  }
}