/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.e.e01;

import drjava.util.Display;
import drjava.util.Displayable;
import drjava.util.Factory;
import prophecy.common.InspectableParent;
import prophecy.common.InspectionNode;

import java.util.ArrayList;
import java.util.List;

/**
 * Parameters are passed on as Watcher<A, B>
 * A: data type
 * B: watch result type
 */
public class WatcherPool<A, B> implements Displayable, InspectableParent {
  //private int maxWatchers = 5;
  private double radius = 0.5;
  private Factory<Watcher<A, B>> watcherFactory;
  private final ArrayList<Watcher<A, B>> watchers = new ArrayList<Watcher<A,B>>();
  private double spawnThreshold = 0.5;
  private boolean exclusive = true;

  public WatcherPool(Factory<Watcher<A, B>> watcherFactory) {
    this.watcherFactory = watcherFactory;
  }

  public WatcherPool(final Factory<Analyzer<A, B>> analyzerFactory, final Metric<B> metric) {
    watcherFactory = new Factory<Watcher<A, B>>() {
      public Watcher<A, B> make() {
        return new Watcher<A,B>(analyzerFactory, metric);
      }
    };
  }

  /** returns min distance */
  public double processData(A a) {
    double minDistance = getMinDistance(a);

    if (decideSpawn(minDistance)) {
      Watcher<A, B> watcher = watcherFactory.make();
      watcher.processData(a);
      watchers.add(watcher);
    }

    return minDistance;
  }

  public double getMinDistance(A a) {
    double minDistance = 1e6f;
    Watcher<A, B> closestWatcher = null;

    for (Watcher<A, B> watcher : watchers) {
      double distance = watcher.getDistanceToLog(a);
      if (!exclusive && distance < radius)
        watcher.processData(a);
      if (closestWatcher == null || distance < minDistance) {
        minDistance = distance;
        closestWatcher = watcher;
      }
    }

    if (exclusive && closestWatcher != null)
      closestWatcher.processData(a);

    return minDistance;
  }

  /** should a new watcher be created? */
  private boolean decideSpawn(double minDistance) {
    return minDistance >= spawnThreshold;
  }

  public int getSize() {
    return watchers.size();
  }

  public void display(Display display) {
    for (int i = 0; i < watchers.size(); i++) {
      Watcher<A, B> watcher = watchers.get(i);
      display.beginSection("Watcher " + (i+1) + " of " + watchers.size());
      watcher.display(display);
      display.endSection();
    }
  }

  public List<InspectionNode> getInspectionChildren() {
    ArrayList<InspectionNode> children = new ArrayList<InspectionNode>();
    for (int i = 0; i < watchers.size(); i++) {
      Watcher<A, B> watcher = watchers.get(i);
      children.add(new InspectionNode("Watcher " + (i+1), watcher));
    }
    return children;
  }

  public double getSpawnThreshold() {
    return spawnThreshold;
  }

  public void setSpawnThreshold(double spawnThreshold) {
    this.spawnThreshold = spawnThreshold;
  }

  public List<B> getResults() {
    List<B> results = new ArrayList<B>();
    for (Watcher<A, B> watcher : watchers) {
      results.add(watcher.getResult());
    }
    return results;
  }

  public double getRadius() {
    return radius;
  }

  public void setRadius(double radius) {
    this.radius = radius;
  }
}