/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.e.e04;

import drjava.util.GUIUtil;
import prophecy.common.Prophecy;
import prophecy.common.Surface;
import prophecy.common.image.ImageSet;
import prophecy.common.image.PixelUtil;
import prophecy.common.image.RGB;
import prophecy.common.image.RGBImage;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;

public class Rects implements StationSpotter {
  private static final double sizeBias = 0.01; // 0.01-0.02 is a good value
  private static final int iterations = 30; // 30 is a good value

  class Rect {
    int x, y, w, h;

    public Rect(int x, int y, int w, int h) {
      this.x = x;
      this.y = y;
      this.w = w;
      this.h = h;
    }

    public boolean legit() {
      return x >= 0 && y >= 0 && x+w <= image.getWidth() && y+h <= image.getHeight();
    }

    public double getQuality() {
      double sum = getColorDeviation();
      //System.out.println("Size: " + w + "*" + h + ", sum: " + sum);
      return (w+h)*sizeBias -sum; // favor larger rects
    }

    public double getColorDeviation() {
      Collection<RGB> pixels = getPixels();
      RGB squareColor = PixelUtil.getAverage(pixels);
      double sum = 0;
      for (RGB pixel : pixels) {
        sum += PixelUtil.getSquaredDiff(pixel, squareColor);
      }
      sum = Math.sqrt(sum)/pixels.size();
      return sum;
    }

    private Collection<RGB> getPixels() {
      ArrayList<RGB> pixels = new ArrayList<RGB>();
      for (int yy = y; yy < y+h; yy++)
        for (int xx = x; xx < x+w; xx++)
          pixels.add(image.getRGB(xx, yy));
      return pixels;
    }
  }

  final int xstep = 8, ystep = 8;
  RGBImage image;
  ArrayList<Rect> rects;

  public boolean stationSpotted(RGBImage image) {
    this.image = image;
    rects = new ArrayList<Rect>();
    for (int y = ystep; y < image.getHeight()-ystep; y += ystep)
      for (int x = xstep; x < image.getWidth()-xstep; x += xstep) {
        Rect rect = new Rect(x, y, 1, 1);
        rect = adjust(rect);
        if (rect != null)
          rects.add(rect);
      }
    return false;
  }

  private Rect adjust(Rect s) {
    for (int i = 0; i < iterations; i++) {
      Prioritizer<Rect> prioritizer = new Prioritizer<Rect>();
      addIfLegit(prioritizer, s);
      addIfLegit(prioritizer, new Rect(s.x, s.y, s.w+1, s.h));
      addIfLegit(prioritizer, new Rect(s.x, s.y, s.w, s.h+1));
      addIfLegit(prioritizer, new Rect(s.x-1, s.y, s.w+1, s.h));
      addIfLegit(prioritizer, new Rect(s.x, s.y-1, s.w, s.h+1));
      s = prioritizer.getBest();
      if (s == null) return s;
    }
    return s;
  }

  private void addIfLegit(Prioritizer<Rect> prioritizer, Rect rect) {
    if (rect.legit())
      prioritizer.add(rect, rect.getQuality());
  }

  public Surface getSurface() {
    final int zoom = 2;
    return new Surface() {
      public void render(int width, int height, Graphics2D g2) {
        g2.setTransform(AffineTransform.getScaleInstance(zoom, zoom));
        if (image != null)
          g2.drawImage(image.getBufferedImage(), 0, 0, image.getWidth(), image.getHeight(), null);
        if (rects != null)
          for (Rect rect : rects) {
            double deviation = rect.getColorDeviation()*1000;
            g2.setColor(PixelUtil.blend(new RGB(Color.yellow), new RGB(Color.blue), deviation)
              .getColor());
            g2.drawRect(rect.x, rect.y, rect.w, rect.h);
          }
      }
    };
  }

  public static void main(String[] args) throws IOException {
    ImageSet images = new ImageSet(Prophecy.dataDir() + "aibopics", "bongo", "bongo_station", "bongo_station2");
    Rects instance = new Rects();
    instance.stationSpotted(images.getImage(1));

    JFrame frame = new JFrame("SpotTheStation");
    frame.setSize(820, 650);
    frame.getContentPane().add(instance.getSurface());
    GUIUtil.showMainFrame(frame);
  }
}