/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.e.e09;

import drjava.util.Display;
import drjava.util.Displayable;
import drjava.util.ToTree;
import drjava.util.Tree;
import prophecy.common.image.RGB;
import prophecy.common.image.RGBImage;

import java.text.DecimalFormat;

public class ColorCube implements Displayable, ToTree {
  private final int graininess;
  private double[][][] stats;

  public ColorCube() {
    this(3);
  }

  public ColorCube(int graininess) {
    this.graininess = graininess;
    stats = new double[graininess][graininess][graininess];
  }

  public void addPixels(RGBImage image) {
    for (int y = 0; y < image.getHeight(); y++)
      for (int x = 0; x < image.getWidth(); x++)
        addPixel(image.getRGB(x, y));
  }

  private void addPixel(RGB rgb) {
    stats[toIndex(rgb.r)][toIndex(rgb.g)][toIndex(rgb.b)]++;
  }

  private int toIndex(float value) {
    return Math.max(0, Math.min(graininess-1, (int) (value*graininess)));
  }

  public void display(Display display) {
    DecimalFormat numberFormat = new DecimalFormat("0000");
    for (int r = 0; r < graininess; r++) {
      display.indent();
      for (int g = 0; g < graininess; g++) {
        for (int b = 0; b < graininess; b++)
          display.put(numberFormat.format(stats[r][g][b])).put(" ");
        display.nl();
      }
      display.unindent();
      display.nl();
    }
  }

  public ColorCube minus(ColorCube cube) {
    ColorCube result = new ColorCube();
    for (int r = 0; r < graininess; r++)
      for (int g = 0; g < graininess; g++)
        for (int b = 0; b < graininess; b++)
          result.stats[r][g][b] = stats[r][g][b] - cube.stats[r][g][b];
    return result;
  }

  public ColorCube denoise() {
    double threshold = getAbsMax() / 10;
    ColorCube result = new ColorCube();
    for (int r = 0; r < graininess; r++)
      for (int g = 0; g < graininess; g++)
        for (int b = 0; b < graininess; b++) {
          double v = stats[r][g][b];
          result.stats[r][g][b] = Math.abs(v) <= threshold ? 0 : v;
        }
    return result;
  }

  public double getAbsMax() {
    double max = 0;
    for (int r = 0; r < graininess; r++)
      for (int g = 0; g < graininess; g++)
        for (int b = 0; b < graininess; b++)
          max = Math.max(max, Math.abs(stats[r][g][b]));
    return max;
  }

  public ColorCube merge(ColorCube cube, Function2<Double, Double, Double> f) {
    ColorCube result = new ColorCube();
    for (int r = 0; r < graininess; r++)
      for (int g = 0; g < graininess; g++)
        for (int b = 0; b < graininess; b++)
          result.stats[r][g][b] = f.get(stats[r][g][b], cube.stats[r][g][b]);
    return result;
  }

  public int[] getDominantCell() {
    int bestR = 0, bestG = 0, bestB = 0;
    for (int r = 0; r < graininess; r++)
      for (int g = 0; g < graininess; g++)
        for (int b = 0; b < graininess; b++) {
          if (stats[r][g][b] > stats[bestR][bestG][bestB]) {
            bestR = r;
            bestG = g;
            bestB = b;
          }
        }
    return new int[] { bestR, bestG, bestB };
  }

  public Tree toTree() {
    return new Tree(getClass()).add("graininess", graininess);
  }
}