/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.e.e10;

import drjava.util.*;
import static drjava.util.TreeUtil.treeToList;
import prophecy.common.Bayes;
import prophecy.common.BayesParams;
import prophecy.common.SubsetTI;
import prophecy.common.image.RGBImage;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.TreeSet;

public class BayesContestant implements E10Contestant, Displayable, ToTree {
  final Bayes bayes;
  private int maxWords;
  private List<ImageVerbalizer> verbalizers = new ArrayList<ImageVerbalizer>();

  public BayesContestant(BayesParams bayesParams, int maxWords, List<ImageVerbalizer> verbalizers) {
    this.verbalizers = verbalizers;
    this.maxWords = maxWords;
    bayes = new Bayes(bayesParams);
  }

  public void setVerbalizers(List<ImageVerbalizer> verbalizers) {
    this.verbalizers = verbalizers;
  }

  public boolean contains(RGBImage image) {
    Collection<String> words = imageToWords(image);
    double p = bayes.getCombinedProbability(words);
    return p < 0.5;
  }

  private Collection<String> imageToWords(RGBImage image) {
    TreeSet<String> words = new TreeSet<String>();
    for (ImageVerbalizer verbalizer : verbalizers) {
      words.addAll(verbalizer.imageToWords(image));
    }
    return words;
  }

  public void train(List<SubsetTI<RGBImage>> data) {
    for (SubsetTI<RGBImage> item : data) {
      Collection<String> words = imageToWords(item.example);
      //System.out.println(item.inSet + " " + StringUtil.join(words, " "));
      if (item.inSet) {
        bayes.addHamCorpus(1);
        for (String word : words)
          bayes.addHamWord(word, 1);
      } else {
        bayes.addSpamCorpus(1);
        for (String word : words)
          bayes.addSpamWord(word, 1);
      }
    }

    if (maxWords != 0)
      bayes.simplify(maxWords);
  }

  public void display(Display display) {
    display.putnl(toString()).put(bayes);
  }

  public String toString() {
    String s = "BayesContestant " + bayes.getParams();
    for (ImageVerbalizer verbalizer : verbalizers) {
      s += " " + ObjectUtil.toNiceString(verbalizer);
    }
    return s + " bayes: " + bayes.tableSizes();
  }

  public Tree toTree() {
    Tree tree = new Tree(getClass())
      .add("bayes", TreeUtil.objectToTree(bayes))
      .add("verbalizers", TreeUtil.listToTree(verbalizers));
    if (maxWords != 0)
      tree.add("maxWords", maxWords);
    return tree;
  }

  public BayesContestant(Tree tree) {
    bayes = tree.getObject("bayes");
    verbalizers = treeToList(tree.get("verbalizers"));
    maxWords = tree.getInt("maxWords", 0);
  }
}