/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.joe.joe04;

import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;

/**
 * The ImageUtilities class is full of handy static methods.
 * <p>
 * <p>
 * @author Bastien Saltel
 *
 * media tracker memory leak fixed by jazz
 */

public class ImageUtilitiesFixed {
	/** The component. */
	private static final Component component = new Component() {};

   /** The mediatracker.associated with the component */
	private static final	MediaTracker mediatracker = new MediaTracker(component);

    /** The id of the image. */
	private static int id = 0;

    /** The width of the image. */
	private	static int width = 0;

    /** The height of the image. */
	private static int height = 0;

	/**
	 * Sets the width of the image.
     * <p>
	 */
	public static void		setWidth(int i)
	{
		width = i;
	}

	/**
	 * Sets the height of the image.
     * <p>
	 */
	public static void		setHeight(int i)
	{
		height = i;
	}

	/**
	 * Returns the width of the image.
     * <p>
	 */
	public static int		getWidth()
	{
		return width;
	}

	/**
	 * Returns the height of the image.
     * <p>
	 */
	public static int		getHeight()
	{
		return height;
	}

	/**
	 * Waits for the given image to load fully. Returns true if
	 * everything goes well or false if there is an error while loading
	 * the image.
     * <p>
	 * @return true if everything goes well or false if there is an error
	 * during loading the image.
	 */
	public static boolean	waitForImage(Image image)
	{
		int		i;

		synchronized (component)
			{
				i = id++;
			}
    try {
        mediatracker.addImage(image, i, width, height);
				mediatracker.waitForID(i);
        if (mediatracker.isErrorID(i))
          return false;
			}
		catch (InterruptedException ie)
			{
				return false;
			}
    finally {
      mediatracker.removeImage(image);
    }
		return true;
	}

	/**
	 * Loads an image from the given binary buffer and will not return until
	 * the image is fully loaded.
	 * <p>
	 * @param buffer The buffer containing the binary data of the image.
	 * @return  The fully loaded image if everything goes well or null if there
	 * is an error while loading.
	 */
	public static Image		blockingLoad(byte[] buffer)
	{
		Image image = Toolkit.getDefaultToolkit().createImage(buffer);
		if (waitForImage(image) == false)
			return null;
		return image;
	}

	/**
	 * Creates a buffered image from the supplied image.
	 * <p>
	 * @param image  The supplied image.
	 * @return  The buffered image.
	 */
	public static BufferedImage makeBufferedImage(Image image)
	{
		if (image == null)
			System.out.println("ImageUtilities l.67 Image null");
		return makeBufferedImage(image, BufferedImage.TYPE_INT_RGB);
	}

	/**
	 * Creates a buffered image from the supplied image and its type.
	 * <p>
	 * @param image  The supplied image.
	 * @param imageType The type of the image.
	 * @return  The buffered image.
	 */
	public static BufferedImage		makeBufferedImage(Image image, int imageType)
	{
		if (waitForImage(image) == false)
			return null;
		if (image == null)
			System.out.println("ImageUtilities l.77 Image null");
		BufferedImage	bufferedImage = new BufferedImage(width, height, imageType);

		Graphics2D g = bufferedImage.createGraphics();
		if (image == null)
			System.out.println("ImageUtilities l.80 Image null");
		g.drawImage(image, null, null);
		return bufferedImage;
	}

	/**
	 * Returns a frame subclass whose the update mathod does not clear the frame's
	 * drawing surface.
	 * <p>
	 * @param name  The frame's title.
	 * @param c  The component whose is used to fit the frame's size.
	 * @return	The generated frame.
	 */
	public static Frame		getNonClearingFrame(String name, Component c)
	{
		final Frame f = new Frame(name)
			{
				public void		update(Graphics g)
				{
					paint(g);
				}
			};
		sizeContainerToComponent(f, c);
		centerFrame(f);
		f.setLayout(new BorderLayout());
		f.add(c, BorderLayout.CENTER);
		f.addWindowListener(new WindowAdapter()
			{
				public void windowClosing(WindowEvent e)
				{
					f.dispose();
				}
			});
		return f;
	}

	/**
	 * Resizes the given container to enclose the preferred size of the given component.
	 * <p>
	 * @param container The container to resize.
	 * @param component The component.
	 */
	public static void		sizeContainerToComponent(Container container, Component component)
	{
		if (container.isDisplayable() == false)
			container.addNotify();
		Insets insets = container.getInsets();
		Dimension size = component.getPreferredSize();
		int		w = insets.left + insets.right + size.width;
		int		h = insets.top + insets.bottom + size.height;
		container.setSize(w, h);
	}

	/**
	 * Places the given frame in the center of the screen.
	 */
	public static void		centerFrame(Frame f)
	{
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension d = f.getSize();
		int		x = (screen.width - d.width) / 2;
		int		y = (screen.height - d.height) / 2;
		f.setLocation(x, y);
	}
}