/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.joe.joe09;

import drjava.util.Log;
import static drjava.util.StringUtil.formatDouble;
import prophecy.audio.a11.SoundChunker;
import prophecy.common.Trigger;
import prophecy.common.sound.Sound;
import prophecy.joe.joe02.RotatingSoundBuffer;
import prophecy.joe.joe02.SoundIn;
import prophecy.joe.joe02.SoundListener;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;

public class WordRecorder implements SoundListener {
  private double minDuration = 0.2, maxDuration = 2.0;
  private boolean buffering = true;
  private SoundBuffer chunkBuffer;
  private WordCollector wordCollector;
  private Trigger bufferProcessedTrigger = new Trigger();
  private double ignoreNextWordCountdown;
  private List<WordRecorderListener> listeners = new ArrayList<WordRecorderListener>();

  // statistics
  private String lastAction;
  private int chunksInSnippet;
  private String ignoreNextWordReason;
  private RotatingSoundBuffer buffer;

  public WordRecorder(SoundIn soundIn, WordCollector wordCollector) {
    this.wordCollector = wordCollector;
    buffer = new RotatingSoundBuffer(0.5);
    soundIn.addListener(buffer);
    buffer.addListener(this);
  }

  public void soundReceived(final Sound sound) {
    SoundChunker chunker = new SoundChunker(sound);
    List<Rectangle> chunks = chunker.findChunks();
    chunksInSnippet = chunks.size();
    boolean touchLeft = !chunks.isEmpty() && chunks.get(0).x == 0;
    if (!touchLeft && buffering) {
      if (chunkBuffer != null)
        chunkFound(chunkBuffer.getSound());
      chunkBuffer = null;
      buffering = false;
    }

    for (Rectangle r : chunks) {
      double sec1 = chunker.toSeconds(r.x);
      double sec2 = chunker.toSeconds(r.x + r.width);
      boolean touchRight = sec2 + 0.01 >= sound.getSeconds();
      Sound chunk = sound.clip(sec1, sec2);
      if (touchRight) {
        if (!buffering) {
          chunkBuffer = new SoundBuffer();
          buffering = true;
        }
        if (chunkBuffer != null) {
          chunkBuffer.append(chunk);
          if (chunkBuffer.getSeconds() > maxDuration)
            chunkBuffer = null; // discard
        }
      } else {
        if (buffering) {
          if (chunkBuffer != null) {
            chunkBuffer.append(chunk);
            chunkFound(chunkBuffer.getSound());
          }
          chunkBuffer = null;
          buffering = false;
        } else
          chunkFound(chunk);
      }
    }

    if (ignoreNextWordCountdown > 0) {
      ignoreNextWordCountdown = Math.max(0, ignoreNextWordCountdown-sound.getSeconds());
      if (ignoreNextWordCountdown == 0)
        Log.surprise("no word came");
    }

    bufferProcessedTrigger.trigger();
  }

  private void chunkFound(Sound chunk) {
    if (chunk.getSeconds() < minDuration) {
      lastAction = "discarded, too short";
      return;
    }

    if (chunk.getSeconds() > maxDuration) {
      lastAction = "discarded, too long";
      return;
    }

    if (ignoreNextWordCountdown > 0) {
      lastAction = "ignored" + (ignoreNextWordReason == null || ignoreNextWordReason.length() == 0
        ? "" : " (" + ignoreNextWordReason + ")");
      ignoreNextWordCountdown = 0;
      return;
    }

    wordFound(chunk);
    lastAction = "word saved (" + formatDouble(chunk.getSeconds(), 1) + "s)";
  }

  private void wordFound(Sound chunk) {
    wordCollector.addWord(chunk);
    for (WordRecorderListener listener : listeners)
      listener.wordRecorded(chunk);
  }

  public Trigger getBufferProcessedTrigger() {
    return bufferProcessedTrigger;
  }

  public String getWordInFlightStats() {
    return (buffering ? (chunkBuffer == null ? "too long, discarding"
      : drjava.util.StringUtil.formatDouble(chunkBuffer.getSeconds(), 1)) : "-");
  }

  public int getChunksInSnippet() {
    return chunksInSnippet;
  }

  public String getLastAction() {
    return lastAction;
  }

  public double getIgnoreNextWordCountdown() {
    return ignoreNextWordCountdown;
  }

  public String getIgnoreNextWordReason() {
    return ignoreNextWordReason;
  }

  public void setIgnoreNextWordCountdown(double seconds, String reason) {
    ignoreNextWordCountdown = seconds;
    ignoreNextWordReason = reason;
  }

  public void addListener(WordRecorderListener listener) {
    listeners.add(listener);
  }

  public float getSampleRate() {
    return buffer.getSampleRate(); 
  }
}