/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.m.m02;

import drjava.util.LetterLayout;
import drjava.util.TextView;
import prophecy.common.gui.SheetApp;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.DateFormat;
import java.util.Date;

public class ClockWatch extends SheetApp {
  private JLabel lClock;
  private StringWatcher stringWatcher;
  private CharsWatcher charsWatcher;
  private TextView analysisView;
  private boolean showSeconds = true;
  private JLabel lPrediction;
  private String clockText;
  private JLabel lAccuracy;
  private PredictionSuccessMeter<String> predictionSuccessMeter;
  private ChangeWatcher<String> changeWatcher;
  private ChangeAwareWatcher<String> changeAwareWatcher;
  private CombinedWatcher<String> combinedWatcher;
  private Timer timer;
  private int frequency = 4;

  public ClockWatch() {
    super("M02: Clock Watch");

    addButton("Restart", new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        initWatchers();
      }
    });

    final JRadioButton rb4Hz = new JRadioButton("4 Hz", true);
    JRadioButton rb10Hz = new JRadioButton("10 Hz");
    ButtonGroup frequencyButtons = new ButtonGroup();
    frequencyButtons.add(rb4Hz);
    frequencyButtons.add(rb10Hz);
    JPanel frequencyPanel = new JPanel(LetterLayout.leftAlignedRow(10));
    frequencyPanel.add(new JLabel("Sample frequency:"));
    frequencyPanel.add(rb4Hz);
    frequencyPanel.add(rb10Hz);
    addComponent(frequencyPanel);
    frequencyButtons.getSelection().addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        setFrequency(rb4Hz.isSelected() ? 4 : 10);
      }
    });

    lClock = makeClockLabel();
    addBlueHeading("Input");
    addComponent(lClock);

    lPrediction = makeClockLabel();
    addBlueHeading("Prediction");
    addComponent(lPrediction);
    lAccuracy = new JLabel(" ", JLabel.CENTER);
    addComponent(lAccuracy);

    addBlueHeading("Analysis");
    analysisView = new TextView();
    analysisView.setMinimumSize(new Dimension(300, 300));
    analysisView.setInset(3);
    addComponent(analysisView);
  }

  private void setFrequency(int freq) {
    if (frequency != freq) {
      frequency = freq;
      timer.setDelay(1000/freq);
      initWatchers();
    }
  }

  private JLabel makeClockLabel() {
    JLabel label = new JLabel(" ", JLabel.CENTER);
    label.setFont(new Font("Courier New", Font.PLAIN, 24));
    return label;
  }

  public void start() {
    showFrame();

    initWatchers();

    timer = new Timer(1000/4, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        step();
      }
    });
    timer.start();
  }

  private void initWatchers() {
    combinedWatcher = new CombinedWatcher<String>();
    stringWatcher = new StringWatcher();
    combinedWatcher.addWatcherFront(stringWatcher);
    changeAwareWatcher = new ChangeAwareWatcher<String>(combinedWatcher);
    changeWatcher = changeAwareWatcher.getChangeWatcher();
    ImprovedRunlengthWatcher boolWatcher = new ImprovedRunlengthWatcher();
    changeWatcher.setBoolWatcher(boolWatcher);
    predictionSuccessMeter = new PredictionSuccessMeter<String>(changeAwareWatcher);
  }

  private void step() {
    Date date = new Date();
    clockText = DateFormat.getTimeInstance().format(date);
    if (!showSeconds)
      clockText = clockText.substring(0, 5);
    lClock.setText("<html><center>" + clockText + "</center></html>");
    watch(clockText);

    //String analysis = technicalAnalysis();
    String analysis = readableAnalysis();
    analysisView.setText(analysis, true);

    predict();
  }

  private String readableAnalysis() {
    Typewriter typewriter = new Typewriter();

    typewriter.addSentence("Change prediction: " + changeWatcher.predict());
    changeAwareWatcher.describeYourself(typewriter, "clock time");
    return typewriter.getHtml();
  }

  private String technicalAnalysis() {
    String analysis = analyzeWatcher(stringWatcher);
    analysis += "<br>";
    if (charsWatcher != null) {
      for (int i = 0; i < charsWatcher.getLength(); i++) {
        StringWatcher watcher = charsWatcher.getCharWatcher(i);
        analysis += "#" + (i + 1) + "/" + charsWatcher.getLength() + "<br>"
          + "Model: " + drjava.util.ObjectUtil.nice(watcher.getModel()) + "<br><br>";
      }
    }
    return analysis;
  }

  private void predict() {
    Prediction<String> prediction = changeAwareWatcher.predict();
    lPrediction.setText("<html><center>" + (prediction == null ? "" : prediction.getValue()) + "</center></html>");

    lAccuracy.setText("Accuracy: " + predictionSuccessMeter.getPercent() + "%");
  }

  private String analyzeWatcher(StringWatcher watcher) {
    String analysis = "Model: " + drjava.util.ObjectUtil.nice(this.stringWatcher.getModel()) + "<br>"
      + "Length: " + drjava.util.ObjectUtil.nice(this.stringWatcher.getLengthModel()) + "<br>";
    return analysis;
  }

  private void watch(String text) {
    changeAwareWatcher.feed(text);
    if (charsWatcher == null) {
      charsWatcher = stringWatcher.makeCharsWatcher();
      combinedWatcher.addWatcherBack(charsWatcher);
    }
    predictionSuccessMeter.feed(text);
  }

  public static void main(String[] args) {
    ClockWatch clockWatch = new ClockWatch();
    clockWatch.asMainApplication();
    clockWatch.start();
  }
}