/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.m.m31;

import drjava.superversion.filesystem.DirectoryContents;
import drjava.superversion.filesystem.MutableDirectory;
import drjava.superversion.filesystem.RealFileSystem;
import drjava.util.GUIUtil;
import static drjava.util.ObjectUtil.nice;
import drjava.util.RealRandomizer;
import drjava.util.StdOutLogger;
import jazz.retroui.RetroUIFrame;
import prophecy.common.comments.TreeServer;
import prophecy.common.comments.Treelet;
import prophecy.common.gui.AutoVMExit;
import prophecy.m.m01.SocketUtil;
import prophecy.m.m26.ProphecyJarMaker;
import prophecy.m.m32.JBLTreelet;

import java.io.*;
import java.net.ServerSocket;

public class JavaBootLoader {
  private String id;
  private RollingScreen screen;
  private MutableDirectory runDir;
  private RandomAccessFile logOut;
  private MutableDirectory mainDir;

  public JavaBootLoader() {
    id = "jbl" + new RealRandomizer().random(1000);
  }

  public static void main(String[] args) {
    AutoVMExit.install();
    JavaBootLoader jbl = new JavaBootLoader();
    jbl.showFrame();
  }

  private void showFrame() {
    RetroUIFrame frame = new RetroUIFrame("Java Boot Loader", 80, 25);
    GUIUtil.showFrame(frame);
    screen = new RollingScreen(frame.getScreen());
    screen.println("Java Boot Loader");
    screen.println("");

    try {
      systemInfo();
    } catch (Throwable e) { error(e); }

    try {
      discoverDirectories();
    } catch (Throwable e) { error(e); }

    try {
      openLog();
    } catch (Throwable e) { error(e); }

    try {
      redirectSystemOutAndErr();
    } catch (Throwable e) { error(e); }

    try {
      screen.print("Socket: ");
      int port = openSocket();
      screen.println("open at port " + port);
    } catch (Throwable e) { error(e); }

    try {
      autostart();
    } catch (Throwable e) { error(e); }
  }

  private void autostart() {
    screen.println("");
    screen.print("Autostart: ");
    MutableDirectory autostartDir = mainDir.getDirectory("autostart");
    if (!autostartDir.exists()) {
      screen.println(autostartDir.getPath() + " not found");
      return;
    }
    screen.println("Scanning " + autostartDir.getPath());

    DirectoryContents contents = autostartDir.getContents(new StdOutLogger());
    for (drjava.superversion.filesystem.File file : contents.getFiles()) {
      if (file.getName().endsWith(".jar")) try {
        runJar(file);
      } catch (Throwable e) { error(e); }
    }
  }

  private void redirectSystemOutAndErr() {
    screen.print("Redirecting system out/err: ");
    final PrintStream realOut = System.out;
    PrintStream out = new PrintStream(new OutputStream() {
      public void write(int b) throws IOException {
        realOut.write(b);
        //screen.print("" + (char) b);
        if (logOut != null) logOut.write(b);
      }
    });
    System.setOut(out);
    final PrintStream realErr = System.err;
    PrintStream err = new PrintStream(new OutputStream() {
      public void write(int b) throws IOException {
        realErr.write(b);
        if (logOut != null) logOut.write(b);
        screen.print("" + (char) b);
      }
    });
    System.setErr(err);
    screen.println("OK");
  }

  private void openLog() throws FileNotFoundException {
    screen.print("Logging to: ");
    File logFile = new File(runDir.getPath(), "log.txt");
    logOut = runDir == null ? null : new RandomAccessFile(logFile, "rw");
    screen.println(logFile.getPath());
  }

  private void discoverDirectories() {
    File mainDirPath = new File("jbl").getAbsoluteFile();
    RealFileSystem fs = new RealFileSystem();
    mainDir = fs.getDirectory(mainDirPath.getPath());
    mainDir.create();
    screen.println("Using main dir: " + mainDirPath);

    screen.print("Using run dir: ");
    MutableDirectory runs = mainDir.getDirectory("runs");
    runs.create();
    //new File(mainDir, "runs").mkdirs();
    runDir = runs.getDirectory("" + System.currentTimeMillis());
    runDir.create();
    screen.println(runDir.getPath());
  }

  private void systemInfo() {
    screen.println("ID: " + id);
    screen.println("Java version: " + System.getProperty("java.version"));
    screen.println("Available processors: " + Runtime.getRuntime().availableProcessors());
    screen.println("Total memory: " + (Runtime.getRuntime().totalMemory()/1024) + "K");
    screen.println("Free memory: " + (Runtime.getRuntime().freeMemory()/1024) + "K");
  }

  private void error(Throwable e) {
    e.printStackTrace();
    screen.println(nice(e));
  }

  private int openSocket() throws IOException {
    ServerSocket serverSocket = SocketUtil.serverSocketAtPortAbove(40000);
    new TreeServer(serverSocket, makeTreelet()).start();
    return serverSocket.getLocalPort();
  }

  protected Treelet makeTreelet() {
    return new JBLTreelet();
  }

  public static class MakeJar {
    public static void main(String[] args) throws IOException {
      ProphecyJarMaker.makeJar("standalone/bootloader.jar", JavaBootLoader.class);
    }
  }

  private void runJar(drjava.superversion.filesystem.File file) {
    screen.println("");
    screen.println("*** " + file.getName());
    // TODO
  }
}