/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.m.m35;

import drjava.util.*;
import drjava.util.Errors;
import static prophecy.m.m15.M15Util.*;
import prophecy.m.m15.Range;
import prophecy.m.m15.SourcePane;
import prophecy.m.m15.Triplet;
import prophecy.m.m15.WhitespaceSplitter;
import prophecy.common.gui.SheetApp;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

public class M35FileAnalyzer extends SheetApp {
  private Outlines srcOutlines;
  private SourcePane analysisPane;
  private Outlines cmdOutlines;
  private List<Cmd> commands = new ArrayList<Cmd>();
  private XTextField tfCmd;
  private XEditorPane srcEditorPane;
  private LivingInput srcLivingInput;
  private LivingInput cmdLivingInput;
  private XEditorPane scriptPane;
  private Config config;

  public M35FileAnalyzer(Config config) {
    super(config.getTitle());

    fillCommands();

    String srcPath = "src/prophecy/m/m13/StringUnfreezer.java";
    if (!new File(srcPath).exists())
      srcPath = "../" + srcPath; // wow this is an ugly hack
    String sourceText = loadFile(srcPath);

    srcEditorPane = new XEditorPane();
    srcEditorPane.setText(sourceText);
    srcOutlines = new Outlines(srcEditorPane, srcEditorPane.getPaintHook());
    JScrollPane scrollPane = new JScrollPane(srcEditorPane);
    scrollPane.setMinimumSize(new Dimension(500, 250));
    addComponent(GUIUtil.withTitle(srcPath, scrollPane));

    //addLabel("Snippet info");
    analysisPane = new SourcePane();
    scrollPane = new JScrollPane(analysisPane);
    GUIUtil.setMinimumHeight(scrollPane, 100);
    addComponent(GUIUtil.withTitle("Snippet info", scrollPane));
    //addScrollable(analysisPane, 100);

    srcLivingInput = new LivingInput(srcOutlines, analysisPane, config);
    fillSrcAssociator();

    scriptPane = new XEditorPane();
    scriptPane.setEditable(false);
    srcLivingInput.getScript().getTrigger().addListener(new Runnable() {
      public void run() {
        updateScriptPane();
      }
    });

    tfCmd = new XTextField();
    cmdOutlines = new Outlines(tfCmd, tfCmd.getPaintHook());
    addComponent(GUIUtil.withLabel("Command:", tfCmd));

    tfCmd.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        runCmd();
      }
    });

    cmdLivingInput = new LivingInput(cmdOutlines, analysisPane, config);
    cmdLivingInput.getAssociator().put("split", "command");
    cmdLivingInput.getScript().add(new Cmd() {
      public String getName() {
        return "split";
      }

      public void run() {
        String sourceText = cmdLivingInput.getText();
        WhitespaceSplitter splitter = new WhitespaceSplitter(sourceText);
        for (Range range : splitter.getRanges()) {
          cmdLivingInput.getOutlines().addOutline(
            new Outline(cmdLivingInput.getOutlines().getTextComponent(), sourceText, range.getStart(), range.getEnd()));
        }
      }
    });

    new PopupMenuHelper() {
      protected void fillMenu(JPopupMenu menu, Point point) {
        JMenuItem miAssoc = new JMenuItem("assoc");
        miAssoc.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            Outline outline = srcOutlines.getActiveOutline();
            if (outline != null)
              cmdOut("snippet(" + outline.getStart() + ", " + outline.getEnd()
                + ", " + quote(outline.getText()) + ") - ");
          }
        });
        menu.add(miAssoc);
      }
    }.install(srcEditorPane);

    showFrame();

    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        tfCmd.requestFocus();
      }
    });
  }

  private void cmdOut(String s) {
    tfCmd.setText(s);
    tfCmd.setCaretPosition(s.length());
    tfCmd.requestFocus();
  }

  private void updateScriptPane() {
    StringBuilder builder = new StringBuilder();
    for (Cmd cmd : srcLivingInput.getScript().getCommands()) {
      builder.append(cmd.getName() + "\n");
    }
    scriptPane.setText(builder.toString());
  }

  public void showFrame() {
    JFrame frame = new JFrame(sheet.getTitle());
    frame.getContentPane().setLayout(new LetterLayout("MMS"));
    frame.getContentPane().add("M", getPanel());
    JScrollPane scrollPane = new JScrollPane(scriptPane);
    scrollPane.setMinimumSize(new Dimension(150, 100));
    frame.getContentPane().add("S", GUIUtil.withInset(GUIUtil.withTitle("Script", scrollPane)));
    frame.pack();
    showFinalFrame(frame);
  }

  private void fillCommands() {
    commands.add(new CmdSplit(this));
  }

  private void runCmd() {
    tfCmd.selectAll();
    String text = tfCmd.getText();
    for (Cmd command : commands) {
      if (command.getName().equals(text)) {
        srcLivingInput.getScript().add(command);
        srcLivingInput.runScript();
        return;
      }
    }
  }

  private String loadFile(String fileName) {
    String sourceText = "";
    try {
      sourceText = FileUtil.loadTextFile(fileName);
    } catch (Throwable e) {
      Errors.add(e);
    }
    return sourceText;
  }

  public void split() {
    String sourceText = srcOutlines.getText();
    WhitespaceSplitter splitter = new WhitespaceSplitter(sourceText);
    for (Range range : splitter.getRanges()) {
      srcOutlines.addOutline(new Outline(srcOutlines.getTextComponent(), sourceText, range.getStart(), range.getEnd()));
    }
  }

  private void fillSrcAssociator() {
    addTriplet(new Triplet(quote("package"), "isA", "keyword"));
    addTriplet(new Triplet(quote("public"), "isA", "keyword"));
  }

  private void addTriplet(Triplet triplet) {
    if (triplet.getPredicate().equals("isA") && isQuoted(triplet.getSubject()))
      srcLivingInput.getAssociator().put(unquote(triplet.getSubject()), triplet.getObject());
  }

  public static void main(String[] args) {
    new M35FileAnalyzer(new Config() {
      public IOutlineAnalyzer makeOutlineAnalyzer() {
        return new OutlineAnalyzer();
      }

      public String getTitle() {
        return "M35: File Analyzer";
      }
    }).asMainApplication();
  }
}