package prophecy.mcos.mcos04;

import drjava.util.*;
import prophecy.common.ClassData;
import drjava.util.Errors;
import prophecy.common.FrozenObject;
import prophecy.mcos.mcos01.OneProcessEnv;
import prophecy.mcos.mcos02.ExampleGuiProcess;
import prophecy.mcos.mcos02.GuiProcess;
import prophecy.mcos.mcos02.ProcessArray;
import prophecy.mcos.mcos03.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

public class InheritableMain implements SystemActions {
  protected OneProcessEnv env;
  private ProcessArray processArray;
  private InternalGuiEnv guiEnv;
  private JFrame frame;
  private TaskBar taskBar;
  private int restartDelay = 500, shutdownDelay = 1000;
  private String version = "";
  private boolean bootScreen;

  public InheritableMain() {
  }

  public void start() {
    frame = new JFrame("Mobile Code OS" + (version.length() == 0 ? "" : " " +version));
    frame.setSize(500, 400);

    showBootScreen();
    //init(false);

    GUIUtil.centerOnScreen(frame);
    GUIUtil.showFrame(frame);
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
  }

  private void showBootScreen() {
    bootScreen = true;
    init1();
    init2();
    taskBar.setMinimalMenu(true);
    boolean loadEnabled = hasSavedSystemState();
    startGuiProcess(new GPBootDialog(new BootDialogListener() {
      public void selectionMade(int selection) {
        exitBootScreen(selection);
      }
    }, loadEnabled));
  }

  private void exitBootScreen(int selection) {
    try {
      softShutdown(); // shutdown boot screen

      bootScreen = false;

      if (selection == 0) {
        init1();
        startDummyProcesses();
        init2();
      } else {
        init1();
        loadSystemState();
        init2();
        //startGuiProcess(new GPMessageBox("System state loaded OK (" + dataLength + " bytes)"));
      }
    } catch (Throwable t) {
      Errors.report(t);
    }
  }

  private boolean hasSavedSystemState() {
    return getSavedSystemState() != null;
  }

  private int loadSystemState() {
    String data = getSavedSystemState();
    FrozenObject frozenObject = new FrozenObject(data);
    Result<Object> result = frozenObject.unfreeze();
    processArray = (ProcessArray) result.get();
    return data.length();
  }

  private void init1() {
    processArray = new ProcessArray();
  }

  private void init2() {
    env = new OneProcessEnv(processArray);
    env.setDefaultIdleTime(0.1);
    env.runInSeparateThread();

    final JDesktopPane desktopPane = new JDesktopPane();
    installWindowPositionFixer(desktopPane);

    // react to F10 key (show system menu)

    KeyEventHelper.install();
    Attachments.set(frame, new KeyEventPostProcessor() {
      /** for some reason, the event reaches us twice, so we ignore one of it */
      public boolean ignoreF10;

      public boolean postProcessKeyEvent(KeyEvent e) {
        //KeyboardFocusManager.getCurrentKeyboardFocusManager().getActiveWindow();
        if (e.getKeyCode() == KeyEvent.VK_F10) {
          e.consume();
          if (ignoreF10)
            ignoreF10 = false;
          else {
            ignoreF10 = true;
            taskBar.showSystemMenu();
          }
          return true;
        } else
          return false;
      }
    }, KeyEventPostProcessor.class);

    guiEnv = new InternalGuiEnv(desktopPane);
    taskBar = new TaskBar(guiEnv, processArray, this);

    taskBar.setAdditionalSystemMenuItems(new MenuFiller() {
      public void fillMenu(JPopupMenu menu) {
        fillStartMenu(menu);
      }
    });

    // make process guis visible (after system state was loaded)
    for (prophecy.mcos.mcos01.Process p : processArray.getProcesses()) {
      if (p instanceof GuiProcess) {
        GuiProcess guiProcess = (GuiProcess) p;
        if (guiProcess.isGuiVisibleOnStartup())
          guiProcess.showGui(guiEnv);
      }
    }

    Container contentPane = frame.getContentPane();
    contentPane.removeAll();
    contentPane.setLayout(
      new LetterLayout("D", "D", "T").setBorder(1).setSpacing(0));
    contentPane.add("D", desktopPane);
    contentPane.add("T", taskBar);
    contentPane.validate();
  }

  private void startDummyProcesses() {
    for (int i = 1; i <= 2; i++) {
      GuiProcess process = new ExampleGuiProcess();
      process.setName("app" + i);
      processArray.add(process);
    }
  }

  private static void installWindowPositionFixer(final JDesktopPane desktopPane) {
    new SwingTimerHelper(new Runnable() {
      public void run() {
        for (JInternalFrame frame : desktopPane.getAllFrames()) {
          int x = frame.getX(), y = frame.getY();
          x = Math.min(x, Math.max(desktopPane.getWidth() - 40, 0));
          y = Math.min(y, Math.max(desktopPane.getHeight() - 10, 0));
          frame.setLocation(x, y);
        }
      }
    }, 1000).installOn(desktopPane);
  }

  public void saveSystemState() {
    HibernaterProcess hibernater = new HibernaterProcess(processArray);
    startProcess(hibernater);
    while (!hibernater.done())
      MultiCoreUtil.sleep(10);
    Result<FrozenObject> result = hibernater.getResult();
    if (result.isSuccess()) {
      System.out.println("<hibernate> " + result);
      String data = result.get().toTree().toString();
      int dataLength = data.length();
      saveSystemState(data);
      startGuiProcess(new GPMessageBox("Session saved! (" + dataLength + " bytes)"));
    } else {
      System.out.println("<hibernate failed>");
      result.getException().printStackTrace();
      startGuiProcess(new GPMessageBox("Session could not be saved"));
    }
  }

  public void startProcess(prophecy.mcos.mcos01.Process process) {
    processArray.start_threadSafe(process);
  }

  private String getSavedSystemState() {
    ClassData classData = ClassData.get(this);
    return classData.getString("system state");
  }

  private void saveSystemState(String data) {
    ClassData classData = ClassData.get(this);
    classData.set("system state", data);
    classData.save();
  }

  public void startGuiProcess(GuiProcess guiProcess) {
    startProcess(guiProcess);
    guiProcess.showGui(guiEnv);
  }

  public void shutdown() {
    Container frame = guiEnv.makeFrame("MCOS is shutting down...");
    guiEnv.setClosable(frame, false);
    frame.setSize(400, 100);
    guiEnv.centerOnScreen(frame);
    guiEnv.showFrame(frame);
    Timer timer = new Timer(shutdownDelay, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        System.exit(0);
      }
    });
    timer.setRepeats(false);
    timer.start();
  }

  public void restart() {
    softShutdown();
    Timer timer = new Timer(restartDelay, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        showBootScreen();
      }
    });
    timer.setRepeats(false);
    timer.start();
  }

  private void softShutdown() {
    env.terminateSeparateThread();
    frame.getContentPane().removeAll();
    frame.getContentPane().repaint();
  }

  public void setVersion(String version) {
    this.version = version;
  }

  protected void fillStartMenu(JPopupMenu menu) {
  }

  public JFrame getFrame() {
    return frame;
  }

  public ProcessArray getProcessArray() {
    return processArray;
  }

  public boolean isBootScreen() {
    return bootScreen;
  }
}