/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.t.t09;

import static drjava.util.ObjectUtil.nice;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.List;
import java.util.regex.Pattern;

public class ServerDiscovery {
  private Soup soup = new Soup();

  public ServerDiscovery() {
  }

  public static void main(String[] args) {
    ServerDiscovery serverDiscovery = new ServerDiscovery();
    serverDiscovery.getSoup().addSoupListener(new SoupListener() {
      public void ingredientAdded(Soup soup, Object ingredient) {
        System.out.println("+" + nice(ingredient));
      }
    });
    serverDiscovery.seed();
  }

  public void seed() {
    soup.add(new Localhost());
    soup.add(new IPsOfThisMachine());
  }

  public Soup getSoup() {
    return soup;
  }

  static class Localhost extends ActiveIngredient {
    public void run() {
      produce(new PossibleHost("localhost"));
    }
  }

  static class IPsOfThisMachine extends ActiveIngredient {
    public void run() {
      /*InetAddress addr;
      try {
        addr = InetAddress.getLocalHost();
      } catch (UnknownHostException e) {
        return;
      }
      String name = addr.getHostName();
      String address = addr.getHostAddress();
      System.out.println("name: " + name + ", address: " + address);*/

      try {
        InetAddress[] addresses = InetAddress.getAllByName("localhost");
        System.out.println("addresses: " + addresses.length);
        for (InetAddress address : addresses) {
          produce(new PossibleHost(address.getHostAddress()));
          produce(new ScanNeighbors(address.getHostAddress()));
        }
      } catch (Exception e) {
        e.printStackTrace();
        return;
      }
    }
  }

  static class ScanNeighbors extends ActiveIngredient {
    private String ip;

    public ScanNeighbors(String ip) {
      this.ip = ip;
    }

    public String toString() {
      return "ScanNeighbors " + ip;
    }

    public void run() {
      String[] parts = ip.split("\\.");
      if (parts.length != 4)
        return;
      int original = new Integer(parts[3]);
      int min = Math.max(1, original-4);
      int max = Math.min(255, original+4);
      for (int i = min; i <= max; i++) {
        if (i != original)
          produce(new PossibleHost(parts[0]+"."+parts[1]+"."+parts[2]+"."+i));
      }
    }
  }

  static class PortScan extends ActiveIngredient {
    private String host;

    PortScan(String host) {
      this.host = host;
    }

    public void run() {
      produce(new PossiblePort(host, 400));
      produce(new PossiblePort(host, 55000));
      //produce(new PossiblePort(host, 401));
    }

    public String toString() {
      return "PortScan " + host;
    }
  }

  static class PossiblePort extends ActiveIngredient {
    private String host;
    private int port;

    public PossiblePort(String host, int port) {
      this.host = host;
      this.port = port;
    }

    public String toString() {
      return "PossiblePort " + host + ":" + port;
    }

    public void run() {
      Socket socket;
      try {
        socket = new Socket(host, port);
      } catch (Throwable e) {
        produce(new CouldNotConnect(host, port, nice(e)));
        return;
      }

      produce(new OpenPortFound(host, port));

      try {
        socket.close();
      } catch (IOException e) {
        //
      }
    }
  }

  public static class CouldNotConnect {
    private String host;
    private int port;
    private String reason;

    public CouldNotConnect(String host, int port, String reason) {
      this.host = host;
      this.port = port;
      this.reason = reason;
    }

    public String toString() {
      return "CouldNotConnect " + host + ":" + port + " (" + reason + ")";
    }
  }

  public static class PossibleHost extends ActiveIngredient {
    public final String host;

    public PossibleHost(String host) {
      this.host = host;
    }

    public String toString() {
      return "PossibleHost " + host;
    }

    public boolean isDuplicate() {
      List<PossibleHost> possibleDuplicates = soup.getIngredients(PossibleHost.class);
      for (PossibleHost possibleDuplicate : possibleDuplicates) {
        if (host.equals(possibleDuplicate.host))
          return true;
      }
      return false;
    }

    public void run() {
      if (!addressSafeToContact(host))
        produce(new UnsafeDestination(host));
      else
        produce(new PortScan(host));
    }
  }

  static class UnsafeDestination {
    private String host;

    public UnsafeDestination(String host) {
      this.host = host;
    }

    public String toString() {
      return "UnsafeDestination " + host;
    }
  }

  public static boolean addressSafeToContact(String address) {
    return isIP(address) &&
      (address.startsWith("127.") || address.startsWith("192.")
      || address.startsWith("10."))
      || address.equals("localhost");
  }

  public static boolean isIP(String address) {
    return Pattern.matches("\\d+\\.\\d+\\.\\d+\\.\\d+", address);
  }
}