/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.video.v04;

import prophecy.common.image.ImageProcessing;
import prophecy.common.image.ImageSurface;
import prophecy.common.image.RGBImage;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;

public class NeighborsSurface extends ImageSurface {
  private RGBImage image;
  private int clipSize = 8, step = clipSize;
  private double minSimilarity = 0;
  private boolean diagonal = true;
  private boolean showImage = false;
  private int discrete = 4;

  public NeighborsSurface(RGBImage image) {
    setForeground(Color.blue);
    setImage(image);
  }

  public void setImage(RGBImage image) {
    this.image = image;
    super.setImage(image);
  }

  public void render(int w, int h, Graphics2D g) {
    if (showImage)
      super.render(w, h, g);
    else {
      g.setColor(Color.black);
      g.fillRect(0, 0, w, h);
    }

    List<Rectangle> clips = ImageProcessing.getMiniclipRects(
      image.getWidth(), image.getHeight(), clipSize, clipSize, step, step);
    for (Rectangle clip : clips) {
      List<Rectangle> neighbors = new ArrayList<Rectangle>();
      addNeighbor(neighbors, clip, -clipSize, 0);
      addNeighbor(neighbors, clip, clipSize, 0);
      addNeighbor(neighbors, clip, 0, -clipSize);
      addNeighbor(neighbors, clip, 0, clipSize);

      if (diagonal) {
        addNeighbor(neighbors, clip, -clipSize, -clipSize);
        addNeighbor(neighbors, clip, -clipSize, clipSize);
        addNeighbor(neighbors, clip, clipSize, -clipSize);
        addNeighbor(neighbors, clip, clipSize, clipSize);
      }

      double[] similarities = new double[neighbors.size()];
      double min = 0, max = 0;
      for (int i = 0; i < neighbors.size(); i++) {
        double sim = similarity(clip, neighbors.get(i));
        similarities[i] = sim;
        if (i == 0 || sim < min) min = sim;
        if (i == 0 || sim > max) max = sim;
      }
      if (minSimilarity < min) min = minSimilarity;
      //System.out.println("min/max: " + min + "/" + max);

      for (int i = 0; i < neighbors.size(); i++) {
        Rectangle neighbor = neighbors.get(i);
        double sim = (similarities[i]-min)/(max-min);
        float alpha = (float) (1-sim);
        if (discrete != 0) alpha = (int) (alpha*discrete)/(float) discrete;
        Color color = /*PixelUtil.blend(Color.black, Color.white, sim).getColor()*/getForeground();
        g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
        g.setColor(color);
        int x1 = clip.x + clip.width / 2;
        int y1 = clip.y + clip.height / 2;
        int x2 = neighbor.x + neighbor.width / 2;
        int y2 = neighbor.y + neighbor.height / 2;

        // rotate 90 deg
        /*int temp = x2-x1;
        x2 = x1+(y2-y1);
        y2 = y1-temp;*/

        g.drawLine(x1, y1, (x1 + x2) / 2, (y1 + y2) / 2);
      }
    }
  }

  private double similarity(Rectangle r1, Rectangle r2) {
    RGBImage clip1 = image.clip(r1), clip2 = image.clip(r2);
    return -ImageProcessing.diff(clip1, clip2).averageBrightness();
  }

  private void addNeighbor(List<Rectangle> neighbors, Rectangle clip, int dx, int dy) {
    Rectangle r = new Rectangle(clip);
    r.translate(dx, dy);
    if (new Rectangle(0, 0, image.getWidth(), image.getHeight()).contains(r))
      neighbors.add(r);
  }

  public void setDiagonal(boolean diagonal) {
    this.diagonal = diagonal;
    repaint();
  }

  public void setMinSimilarity(double minSimilarity) {
    this.minSimilarity = minSimilarity;
    repaint();
  }

  public double getMinSimilarity() {
    return minSimilarity;
  }

  public boolean getDiagonal() {
    return diagonal;
  }

  public boolean getShowImage() {
    return showImage;
  }

  public void setShowImage(boolean showImage) {
    this.showImage = showImage;
    repaint();
  }
}