/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.video.v05;

import drjava.util.StringUtil;
import prophecy.common.image.ImageProcessing;
import prophecy.common.image.RGBImage;

import java.awt.*;
import java.util.ArrayList;
import java.util.List;

public class Neighbors {
  private RGBImage image;
  private int clipSize = 8, step = clipSize;
  private Point[] directions;
  private double tolerance = 0.15;
  private Clip[][] clips;

  public int getClipSize() {
    return clipSize;
  }

  public Clip getClip(int col, int row) {
    if (row < 0 || row >= getRows()) return null;
    if (col < 0 || col >= getCols()) return null;
    return clips[row][col];
  }

  public int getCols() {
    return image.getWidth()/clipSize; 
  }

  public int getRows() {
    return image.getHeight()/clipSize;
  }

  public class Clip {
    private Rectangle rect;
    private double[] neighborDiffs = new double[8];

    public Clip(Rectangle rect) {
      this.rect = rect;
    }

    public Rectangle getNeighbor(int direction) {
      Rectangle r = new Rectangle(rect);
      r.translate(directions[direction].x, directions[direction].y);
      if (new Rectangle(0, 0, image.getWidth(), image.getHeight()).contains(r))
        return r;
      else
        return null;
    }

    public Rectangle getRect() {
      return rect;
    }

    public double[] getNeighborDiffs() {
      return neighborDiffs;
    }

    public String toString() {
      List<String> list = new ArrayList<String>();
      for (double diff : neighborDiffs) {
        list.add(StringUtil.formatDouble(diff, 1));
      }
      return "Clip(" + StringUtil.join(", ", list) + ")";
    }
  }

  public Neighbors(RGBImage image) {
    this.image = image;
    // NW N NE E SE S SW W
    directions = new Point[] {
      new Point(-clipSize, -clipSize),
      new Point(0, -clipSize),
      new Point(clipSize, -clipSize),
      new Point(clipSize, 0),
      new Point(clipSize, clipSize),
      new Point(0, clipSize),
      new Point(-clipSize, clipSize),
      new Point(-clipSize, 0)
    };
    calc();
  }

  public RGBImage getImage() {
    return image;
  }

  public void calc() {
    List<Rectangle> clips = ImageProcessing.getMiniclipRects(
      image.getWidth(), image.getHeight(), clipSize, clipSize, step, step);
    this.clips = new Clip[image.getHeight()/clipSize][image.getWidth()/clipSize];
    for (Rectangle clip : clips) {
      List<Rectangle> neighbors = new ArrayList<Rectangle>();
      addNeighbor(neighbors, clip, -clipSize, -clipSize);
      addNeighbor(neighbors, clip, 0, -clipSize);
      addNeighbor(neighbors, clip, clipSize, -clipSize);
      addNeighbor(neighbors, clip, clipSize, 0);
      addNeighbor(neighbors, clip, clipSize, clipSize);
      addNeighbor(neighbors, clip, 0, clipSize);
      addNeighbor(neighbors, clip, -clipSize, clipSize);
      addNeighbor(neighbors, clip, -clipSize, 0);

      Clip theClip = new Clip(clip);
      double min = 0, max = 0;
      for (int i = 0; i < neighbors.size(); i++) {
        double diff = diff(clip, neighbors.get(i));
        theClip.neighborDiffs[i] = diff;
        if (i == 0 || diff < min) min = diff;
        if (i == 0 || diff > max) max = diff;
      }
      if (max < tolerance) max = tolerance;
      //System.out.println("min/max: " + min + "/" + max);

      for (int i = 0; i < neighbors.size(); i++)
        theClip.neighborDiffs[i] = (theClip.neighborDiffs[i]-min)/(max-min);

      this.clips[clip.y/clipSize][clip.x/clipSize] = theClip;
    }
  }

  private double diff(Rectangle r1, Rectangle r2) {
    RGBImage clip1 = image.clip(r1), clip2 = image.clip(r2);
    return ImageProcessing.diff(clip1, clip2).averageBrightness();
  }

  private void addNeighbor(List<Rectangle> neighbors, Rectangle clip, int dx, int dy) {
    Rectangle r = new Rectangle(clip);
    r.translate(dx, dy);
    if (new Rectangle(0, 0, image.getWidth(), image.getHeight()).contains(r))
      neighbors.add(r);
  }

  public Clip[][] getClips() {
    return clips;
  }

  public Point[] getDirections() {
    return directions;
  }
}