/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package prophecy.video.v09;

import drjava.util.Function;
import drjava.util.GUIUtil;
import static drjava.util.GUIUtil.withLabel;
import prophecy.common.gui.Sheet;
import prophecy.common.image.*;
import prophecy.video.v01.ImageMeta;
import prophecy.video.v01.ImageMetaSurface;
import prophecy.video.v03.Histogram;
import prophecy.video.v08.Naive3;
import prophecy.common.gui.SheetApp;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;

public class V09BlurryVision extends SheetApp {
  private final boolean showHistogram = false;
  private BWImage image;
  private ImageSurface originalSurface;
  private ImageSurface blurredSurface;
  private JSlider blurXSlider, blurYSlider;
  private BWImage blurredImage;
  private HistogramSurface histogramSurface;
  private BWImage bwImage;
  private Histogram histogram;
  private ImageMetaSurface bwSurface;
  private JLabel blobsLabel;
  private JSlider brightnessSlider;

  public V09BlurryVision() {
    super("V09: Blurry Vision");

    sheet.getSheetLayout().setBorder(20);

    image = RGBImage.load("data/v07/typewriter.jpg").toBW();
    originalSurface = new ImageSurface(image);
    addBlueHeading("Original scan");
    sheet.addComponent(originalSurface);
    sheet.addButton("Load image...", new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        originalSurface.loadImage();
        image = new RGBImage(originalSurface.getImage()).toBW();
        update();
      }
    });

    blurXSlider = makeSlider(20, 30);
    blurYSlider = makeSlider(8, 30);

    addBlueHeading("Blurred image");
    blurredSurface = new ImageSurface();
    sheet.addComponent(blurredSurface);
    sheet.addComponent(GUIUtil.withLabel("Horizontal blur", blurXSlider));
    sheet.addComponent(GUIUtil.withLabel("Vertical blur", blurYSlider));

    ChangeListener changeListener = new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        update();
      }
    };
    blurXSlider.addChangeListener(changeListener);
    blurYSlider.addChangeListener(changeListener);

    if (showHistogram) {
      addBlueHeading("Histogram");
      histogramSurface = new HistogramSurface(new Histogram(new double[] {0}));
      histogramSurface.setMinimumSize(new Dimension(150, 50));
      sheet.addComponent(histogramSurface);
    }

    brightnessSlider = makeSlider(10, 100);
    brightnessSlider.addChangeListener(changeListener);
    sheet.addComponent(withLabel("Brightness", brightnessSlider));

    blobsLabel = addBlueHeading(" ");
    RGBImage dummyImage = new RGBImage(1, 1, new int[] {0});
    bwSurface = new ImageMetaSurface(dummyImage, new ImageMeta());
    sheet.addComponent(bwSurface);

    update();

    showFrame();
  }

  private void update() {
    makeBlurredImage_blocky();
    blurredSurface.setImage(blurredImage);
    makeHistogram();
    if (histogramSurface != null)
      histogramSurface.setHistogram(histogram);
    makeBWImage();
    List<Rectangle> blobs = findBlobs(bwImage);
    ImageMeta imageMeta = new ImageMeta("", blobs);
    /*for (Rectangle blob : blobs) {
      System.out.println("blob: " + blob);
    }
    System.out.println("");*/
    bwSurface.setImage(image);
    bwSurface.setMeta(imageMeta);
    blobsLabel.setText(
      blobs.size() + " thing" + (blobs.size() == 1 ? "" : "s")
      + " found");
  }

  private JSlider makeSlider(int initialValue, int max) {
    JSlider blurSlider = new JSlider(JSlider.HORIZONTAL, 0, max, initialValue);
    blurSlider.setMajorTickSpacing(5);
    blurSlider.setMinorTickSpacing(1);
    blurSlider.setPaintTicks(true);
    blurSlider.setPaintLabels(true);
    return blurSlider;
  }

  private void makeBWImage() {
    double threshold = histogram.getThresholdBrightness(/*0.5*/0.8);
    bwImage = ImageProcessing.threshold(blurredImage, threshold);
  }

  private Histogram makeHistogram() {
    histogram = new Histogram(blurredImage);
    //histogram = histogram.integrate();
    histogram = histogram.map(new Function<Double, Double>() {
      public Double get(Double x) {
        return Math.max(0, Math.log10(x));
      }
    });
    return histogram;
  }

  private List<Rectangle> findSquares(RGBImage image) {
    double threshold = 0.8;
    return Naive3.findObjectsForThreshold(image, threshold);
  }

  private List<Rectangle> findBlobs(BWImage image) {
    return new BlobFinder().findBlobs(image);
  }

  private void makeBlurredImage_blocky() {
    int fX = Math.max(1, blurXSlider.getValue());
    int fY = Math.max(1, blurYSlider.getValue());
    blurredImage = BlockyBlur.blur(image, fX, fY, brightnessSlider.getValue()*0.01);
  }

  private void makeBlurredImage_slow() {
    int fX = Math.max(1, blurXSlider.getValue());
    int fY = Math.max(1, blurYSlider.getValue());
    PixelMixer pixelMixer;
    if (fX*fY < 4)
      pixelMixer = new MinPixelMixer();
    else
      pixelMixer = new HistogramPixelMixer(0.1);
    int w = image.getWidth();
    int h = image.getHeight();
    float[] pixels = new float[w*h];
    for (int y = 0; y < h; y++)
      for (int x = 0; x < w; x++) {
        Rectangle r = new Rectangle(x-fX/2, y-fY/2, fX, fY);
        pixels[y*w+x] = pixelMixer.mix(image.clip(r));
      }

    blurredImage = new BWImage(w, h, pixels);
  }

  public static void main(String[] args) {
    new V09BlurryVision().asMainApplication();
  }

  public Sheet getSheet() {
    return sheet;
  }
}