package prophecy.video.v13;

import drjava.util.*;
import prophecy.common.ClassData;
import drjava.util.Errors;
import prophecy.common.PersistentTree;
import prophecy.common.gui.PopDownButton;
import prophecy.common.image.ImageProcessing;
import prophecy.common.image.ImageSurface;
import prophecy.common.image.RGB;
import prophecy.common.image.RGBImage;
import prophecy.m.m15.Associator;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;

public class LiveScreenshot implements Updatable {
  private RGBImage image;
  private ImageSurface surface;
  private JFrame frame;
  private ScreenshotAnalysis analysis = new ScreenshotAnalysis();
  private JLabel lblDominantColor;
  private int updateDelay = 1500; // this is in ms
  private Associator associator = new Associator();
  private JLabel lblGuessedLabel;
  private PersistentTree classDataAssociator;
  private boolean showTiles;
  private ClassData classData;
  private TileShape activeShape;

  public void showFrame() throws AWTException {
    loadData();

    image = makeScreenshot();
    RGBImage rgbImage = resizeScreenshot(image);

    surface = new ImageSurface(rgbImage) {
      public void render(int w, int h, Graphics2D g) {
        super.render(w, h, g);
        renderSurfaceOverlays(g);
      }
    };

    JPanel infoPanel = new JPanel(LetterLayout.leftAlignedRow().setBorder(0).setSpacing(5));
    lblDominantColor = new JLabel(" ");
    lblDominantColor.setOpaque(true);
    lblDominantColor.setMinimumSize(new Dimension(20, 5));
    infoPanel.add(new JLabel("Dominant color:"));
    infoPanel.add(lblDominantColor);

    JPanel labelPanel = new JPanel(LetterLayout.leftAlignedRow().setBorder(0).setSpacing(5));
    labelPanel.add(new JLabel("Guessed label:"));
    lblGuessedLabel = new JLabel();
    labelPanel.add(lblGuessedLabel);
    final JButton btnEnterLabel = new JButton("Enter label...");
    //btnEnterLabel.setMargin(new Insets(4, 4, 4, 4));
    btnEnterLabel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String key = makeAnalysisKey();
        String oldLabel = associator.get(key);
        String label = JOptionPane.showInputDialog(btnEnterLabel,
          "Please enter a label for this screenshot", oldLabel == null ? "" : oldLabel);
        if (label != null) {
          associator.put(key, label);
          saveData();
        }
      }
    });
    labelPanel.add(btnEnterLabel);
    labelPanel.add(new PopDownButton() {
      protected void fillMenu(JPopupMenu menu) {
        JMenuItem miShowTiles = new JMenuItem(showTiles ? "Hide tiles" : "Show tiles");
        miShowTiles.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            showTiles = !showTiles;
            saveData();
          }
        });
        menu.add(miShowTiles);

        menu.addSeparator();

        JMenuItem miDeleteLabels = new JMenuItem("Delete all labels");
        miDeleteLabels.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            deleteLabels();
          }
        });
        menu.add(miDeleteLabels);
      }
    });

    surface.addMouseMotionListener(new MouseMotionListener() {
      public void mouseDragged(MouseEvent e) {
      }

      public void mouseMoved(MouseEvent e) {
        surfaceMouseMove(e);
      }
    });

    surface.addMouseListener(new MouseAdapter() {
      public void mouseEntered(MouseEvent e) {
        surfaceMouseMove(e);
      }

      public void mouseExited(MouseEvent e) {
        activateShape(null);
      }
    });

    frame = new JFrame();
    frame.getContentPane().setLayout(new LetterLayout("I", "L", "S", "S").setBorder(3).setSpacing(3));
    frame.getContentPane().add("S", surface);
    frame.getContentPane().add("I", infoPanel);
    frame.getContentPane().add("L", labelPanel);
    frame.pack();
    frame.setResizable(false);
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
    frame.setLocation(screenSize.width-frame.getWidth()-8, screenSize.height-frame.getHeight()-40);
    frame.setAlwaysOnTop(true);
    frame.setVisible(true);

    newImage(image);

    new SwingTimerHelper(this, updateDelay).installOn(surface);
  }

  private void surfaceMouseMove(MouseEvent e) {
    TilesAnalysis tilesAnalysis = analysis.getTilesAnalysis();
    if (tilesAnalysis == null) return;
    Tiles tiles = tilesAnalysis.getTiles();
    int x = e.getX()/tiles.getTileSize(), y = e.getY()/tiles.getTileSize();
    if (x >= 0 && y >= 0 && x < tiles.getCols() && y < tiles.getRows())
      activateShape(tilesAnalysis.getShape(x, y));
    else
      activateShape(null);
  }

  private void renderSurfaceOverlays(Graphics2D g) {
    if (activeShape != null) {
      int tileSize = analysis.getTilesAnalysis().getTiles().getTileSize();
      java.util.List<Point> tiles = activeShape.getTiles();
      g.setColor(Color.gray);
      for (Point p : tiles)
        g.drawRect(p.x*tileSize, p.y*tileSize, tileSize-1, tileSize-1);
    }
  }

  private void activateShape(TileShape shape) {
    if (shape != activeShape) {
      activeShape = shape;
      surface.repaint();
    }
  }

  private void loadData() {
    classData = ClassData.get(this);
    classDataAssociator = classData.subTree("associator");
    associator.fromTree(classDataAssociator);
    showTiles = classData.getBool("showTiles", showTiles);
  }

  private void saveData() {
    classDataAssociator.set(associator.toTree());
    classData.setBool("showTiles", showTiles);
    classData.save();
  }

  private void deleteLabels() {
    associator.clear();
    saveData();
  }

  public void update() {
    try {
      RGBImage rgbImage = makeScreenshot();
      newImage(rgbImage);
    } catch (Throwable t) {
      Errors.add(t);
    }
  }

  private void newImage(RGBImage image) {
    this.image = image;
    showImage(image);

    analysis.newImage(image);
    double percentage = analysis.getDifferencePercentage();
    String info = (int) percentage + "% movement";
    RGB mostFrequentColor = analysis.getDominantColor();
    //info += " " + Integer.toHexString(mostFrequentColor.asInt());
    frame.setTitle(info + " | Tinybrain Live Screenshot");
    lblDominantColor.setBackground(mostFrequentColor.getColor());

    String label = associator.get(makeAnalysisKey());
    lblGuessedLabel.setText(label == null ? "?" : label);
  }

  private String makeAnalysisKey() {
    return analysis.getDominantColor().getHexString();
  }

  private void showImage(RGBImage image) {
    image = resizeScreenshot(image);
    enhanceScreenshot(image);
    surface.setImage(image);
  }

  private void enhanceScreenshot(RGBImage image) {
    if (showTiles)
      ScreenshotEnhancement.paintDominantColorBoxes(image, 8, 8);
  }

  private static RGBImage resizeScreenshot(RGBImage rgbImage) {
    int newWidth = 256;
    int newHeight = (int) ((double) newWidth/rgbImage.getWidth()*rgbImage.getHeight());
    rgbImage = ImageProcessing.resize(rgbImage, newWidth, newHeight, new AvgPixelMixer());
    return rgbImage;
  }

  private static RGBImage makeScreenshot() throws AWTException {
    Robot robot = new Robot();
    Rectangle rect = new Rectangle(new Point(0, 0), Toolkit.getDefaultToolkit().getScreenSize());
    BufferedImage image = robot.createScreenCapture(rect);
    return new RGBImage(image);
  }
}
