/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package sorttreetable;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;

import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.table.DefaultTableModel;
import javax.swing.tree.TreePath;

/***
 * Original code by Philip Milne and Scott Violet for Sun Microsystems<br/>
 * TreeTable Example 2 can be found here:<br/>
 * <a href="http://java.sun.com/products/jfc/tsc/articles/treetable2/">
 * http://java.sun.com/products/jfc/tsc/articles/treetable2/
 * </a><br/>
  *<br/>
 * Original code by Scott Violet for Sun Microsystems<br/>
 * TreeTable Example 3 (Editable Bookmarks) can be found here:<br/>
 * <a href="http://java.sun.com/products/jfc/tsc/articles/bookmarks/">
 * http://java.sun.com/products/jfc/tsc/articles/bookmarks/
 * </a><br/>
 * <br/>
 * Original code by Claude Duguay for "Java Pro"<br/>
 * JSortTable can be found here:<br/>
 * <a href="http://www.fawcette.com/javapro/2002_08/magazine/columns/visualcomponents/default_pf.aspx">
 * http://www.fawcette.com/javapro/2002_08/magazine/columns/visualcomponents/default_pf.aspx
 * </a><br/>
 * <br/>
 * Combines the TreeTable and the SortTable models.<br/>
 * <br/>
 * A SortTreeTableModel has a node-structure, starting with the 'root' node. 
 * It also has a comparator which is used for sorting the nodes. By default, this
 * is the DefaultSortTableComparator, which compares the userObjects contained in
 * the TreeTableNodes.<br/>
 * <br/>
 * Extending this abstract class is as easy as filling in the blank methods.<br/>
 *
 * @author Claude Duguay
 * @author Philip Milne
 * @author Scott Violet
 * @author Vincent Vollers
 * */
public abstract class AbstractSortTreeTableModel
	extends DefaultTableModel
	implements SortTableModel, TreeTableModel {
	/** Names of the columns, used for the TableModel getColumnName method. */
	private String[] columnNames;
	/**
	 * Method names used to determine a particular value. Used for the
	 * TableModel method getValueAt.
	 */
	private String[] methodNames;
	/**
	 * Setter method names, used to set a particular value. Used for the
	 * TableModel method setValueAt. A null entry, or array, indicates the
	 * column is not editable.
	 */
	private String[] setterMethodNames;
	/** Column classes, used for the TableModel method getColumnClass. */
	private Class[] cTypes;
	/***
	 * The root of the tree.
	 */
	protected TreeTableNode root;
	
	/***
	 * The comparator used for sorting the tree.
	 */
	protected SortTableComparator comparator;

	/***
	 * Creates a new DefaultSortTreeTableModel.
	 * call with super() from an extending class.
	 * 
	 * Initializes the default Comparator and the root node.
	 */
	public AbstractSortTreeTableModel(
			String[] columnNames,
			String[] getterMethodNames,
			String[] setterMethodNames,
			Class[] cTypes) {
		comparator 	= new DefaultSortTableComparator();
		root 		= new TreeTableNode();
		this.columnNames = columnNames;
		this.methodNames = getterMethodNames;
		this.setterMethodNames = setterMethodNames;
		this.cTypes = cTypes;
	}

	/***
	 * This default implementation of isSortable always
	 * returns true.
	 */
	public boolean isSortable(int col) {
		return true;
	}

	/***
	 * Sorts the model by the give column and order.
	 */
	public void sortColumn(int col, boolean ascending) {
		comparator.setAscending(ascending);
		comparator.setColumn(col);
		root.sort(comparator);
	}

	/***
	 * Returns the comparator used to sort the model.
	 */
	public SortTableComparator getComparator() {
		return comparator;
	}
	
	/***
	 * Sets the comparator used to sort the model.
	 */
	public void setComparator(SortTableComparator sorter) {
		comparator = sorter;
	}

	/***
	 * Returns the root node.
	 */
	public Object getRoot() {
		return root;
	}

	/***
	 * Overwrite if you are going to user editors in the JTree.
	 * 
	 * The default implementation does nothing (dummy method).
	 * @see javax.swing.tree.DefaultTreeModel#valueForPathChanged(javax.swing.tree.TreePath, java.lang.Object)
	 */
	public void valueForPathChanged(TreePath path, Object newValue) {
	}

	/***
	 * Returns the index of the child node in the parent node.
	 * 
	 * Although the method expects Objects because of the TreeModel contract,
	 * it assumes both parent and child are TreeTableNodes.
	 */
	public int getIndexOfChild(Object parent, Object child) {
		if (parent == null || child == null) {
			return -1;
		}
		return ((TreeTableNode) parent).indexOf(child);
	}

	/***
	 * Adds a listener for the TreeModelEvent posted after the tree changes.
	 */
	public void addTreeModelListener(TreeModelListener l) {
		listenerList.add(TreeModelListener.class, l);
	}
	
	/***
	 * Removes a listener previously added with addTreeModelListener().
	 */
	public void removeTreeModelListener(TreeModelListener l) {
		listenerList.remove(TreeModelListener.class, l);
	}

	/***
	 * Notify all listeners that have registered interest for notification on
	 * this event type. The event instance is lazily created using the
	 * parameters passed into the fire method.
	 * 
	 * @see EventListenerList
	 */
	public void fireTreeNodesChanged(
		Object source,
		Object[] path,
		int[] childIndices,
		Object[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeNodesChanged(e);
			}
		}
	}

	/***
	 * Notify all listeners that have registered interest for notification on
	 * this event type. The event instance is lazily created using the
	 * parameters passed into the fire method.
	 * 
	 * @see EventListenerList
	 */
	public void fireTreeNodesInserted(
		Object source,
		Object[] path,
		int[] childIndices,
		Object[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeNodesInserted(e);
			}
		}
	}

	/***
	 * Notify all listeners that have registered interest for notification on
	 * this event type. The event instance is lazily created using the
	 * parameters passed into the fire method.
	 * 
	 * @see EventListenerList
	 */
	public void fireTreeNodesRemoved(
		Object source,
		Object[] path,
		int[] childIndices,
		Object[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeNodesRemoved(e);
			}
		}
	}

	/***
	 * Notify all listeners that have registered interest for notification on
	 * this event type. The event instance is lazily created using the
	 * parameters passed into the fire method.
	 * 
	 * @see EventListenerList
	 */
	public void fireTreeStructureChanged(
		Object source,
		Object[] path,
		int[] childIndices,
		Object[] children) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		TreeModelEvent e = null;
		// Process the listeners last to first, notifying
		// those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == TreeModelListener.class) {
				// Lazily create the event:
				if (e == null)
					e =
						new TreeModelEvent(
							source,
							path,
							childIndices,
							children);
				((TreeModelListener) listeners[i + 1]).treeStructureChanged(e);
			}
		}
	}

	/***
	 * Returns the child at index 'num' of the given node.
	 * 
	 * Although the method expects an Object because of the TreeModel contract,
	 * in reality it assumes the given node is a 'TreeTableNode'.
	 * @see javax.swing.tree.TreeModel#getChild(java.lang.Object, int)
	 */
	public Object getChild(Object node, int num) {
		return ((TreeTableNode) node).getChildAt(num);
	}

	/***
	 * Returns the number of children a given node has.
	 *
	 * Although the method expects an Object because of the TreeModel contract,
	 * in reality it assumes the given node is a 'TreeTableNode'.
	 * @see javax.swing.tree.TreeModel#getChildCount(java.lang.Object)
	 */
	public int getChildCount(Object node) {
		return ((TreeTableNode) node).getChildCount();
	}

	/***
	 * Returns true when the given node is a 'leaf'.
	 *
	 * Although the method expects an Object because of the TreeModel contract,
	 * in reality it assumes the given node is a 'TreeTableNode'.
	 * @see javax.swing.tree.TreeModel#isLeaf(java.lang.Object)
	 */
	public boolean isLeaf(Object node) {
		return ((TreeTableNode) node).isLeaf();
	}
	
	/**
		 * Returns the number of column names passed into the constructor.
		 */
		public int getColumnCount() {
			return columnNames.length;
		}

		/**
		 * Returns the column name passed into the constructor.
		 */
		public String getColumnName(int column) {
			if (cTypes == null || column < 0 || column >= cTypes.length) {
				return null;
			}
			return columnNames[column];
		}

		/**
		 * Returns the column class for column <code>column</code>. This is set
		 * in the constructor.
		 */
		public Class getColumnClass(int column) {
			if (cTypes == null || column < 0 || column >= cTypes.length) {
				return null;
			}
			return cTypes[column];
		}

		/**
		 * Returns the value for the column <code>column</code> and object <code>node</code>.
		 * The return value is determined by invoking the method specified in
		 * constructor for the passed in column.
		 */
		public Object getValueAt(Object node, int column) {
			try {
				Object nodeObject = ((TreeTableNode) node).getUserObject();
				if(nodeObject != null) {
					Method method =
					nodeObject.getClass().getMethod(methodNames[column], (Class[]) null);
					if (method != null) {
						return method.invoke(nodeObject, (Object[]) null);
					}
				}
			} catch (Throwable th) {
			}

			return null;
		}

		/**
		 * Returns true if there is a setter method name for column <code>column</code>.
		 * This is set in the constructor.
		 */
		public boolean isCellEditable(Object node, int column) {
			return (setterMethodNames != null && setterMethodNames[column] != null);
		}

		/**
		 * Sets the value to <code>aValue</code> for the object <code>node</code>
		 * in column <code>column</code>. This is done by using the setter
		 * method name, and coercing the passed in value to the specified type.
		 */
		// Note: This looks up the methods each time! This is rather inefficient;
		// it should really be changed to cache matching methods/constructors
		// based on <code>node</code>'s class, and <code>aValue</code>'s class.
		public void setValueAt(Object aValue, Object node, int column) {
			boolean found = false;
			try {
				// We have to search through all the methods since the
				// types may not match up.
				Object nodeObject = ((TreeTableNode) node).getUserObject();
				
				Method[] methods = nodeObject.getClass().getMethods();

				for (int counter = methods.length - 1; counter >= 0; counter--) {
					if (methods[counter]
						.getName()
						.equals(setterMethodNames[column])
						&& methods[counter].getParameterTypes() != null
						&& methods[counter].getParameterTypes().length == 1) {
						// We found a matching method
						Class param = methods[counter].getParameterTypes()[0];
						if (!param.isInstance(aValue)) {
							// Yes, we can use the value passed in directly,
							// no coercision is necessary!
							if (aValue instanceof String
								&& ((String) aValue).length() == 0) {
								// Assume an empty string is null, this is
								// probably bogus for here.
								aValue = null;
							} else {
								// Have to attempt some sort of coercision.
								// See if the expected parameter type has
								// a constructor that takes a String.
								Constructor cs =
									param.getConstructor(
										new Class[] { String.class });
								if (cs != null) {
									aValue =
										cs.newInstance(new Object[] { aValue });
								} else {
									aValue = null;
								}
							}
						}
						// null either means it was an empty string, or there
						// was no translation. Could potentially deal with these
						// differently.
						methods[counter].invoke(nodeObject, new Object[] { aValue });
						found = true;
						break;
					}
				}
			} catch (Throwable th) {
				System.out.println("exception: " + th);
			}
			if (found) {
				// The value changed, fire an event to notify listeners.
				TreeTableNode parent = (TreeTableNode) ((TreeTableNode) node).getParent();
				fireTreeNodesChanged(
					this,
					getPathToRoot(parent),
					new int[] { getIndexOfChild(parent, node)},
					new Object[] { node });
			}
		}

		/**
		 * Builds the parents of the node up to and including the root node, where
		 * the original node is the last element in the returned array. The length
		 * of the returned array gives the node's depth in the tree.
		 * 
		 * @param aNode
		 *            the TreeNode to get the path for
		 * @param an
		 *            array of TreeNodes giving the path from the root to the
		 *            specified node.
		 */
		public TreeTableNode[] getPathToRoot(TreeTableNode aNode) {
			return getPathToRoot(aNode, 0);
		}

		/**
		 * Builds the parents of the node up to and including the root node, where
		 * the original node is the last element in the returned array. The length
		 * of the returned array gives the node's depth in the tree.
		 * 
		 * @param aNode
		 *            the TreeNode to get the path for
		 * @param depth
		 *            an int giving the number of steps already taken towards the
		 *            root (on recursive calls), used to size the returned array
		 * @return an array of TreeNodes giving the path from the root to the
		 *         specified node
		 */
		private TreeTableNode[] getPathToRoot(TreeTableNode aNode, int depth) {
			TreeTableNode[] retNodes;
			// This method recurses, traversing towards the root in order
			// size the array. On the way back, it fills in the nodes,
			// starting from the root and working back to the original node.

			/*
			 * Check for null, in case someone passed in a null node, or
			 */
			if (aNode == null) {
				if (depth == 0)
					return null;
				else
					retNodes = new TreeTableNode[depth];
			} else {
				depth++;
				if (aNode == root)
					retNodes = new TreeTableNode[depth];
				else
					retNodes = getPathToRoot((TreeTableNode) aNode.getParent(), depth);
				retNodes[retNodes.length - depth] = aNode;
			}
			return retNodes;
		}
}