/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package sorttreetable;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.TableModelEvent;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumnModel;

/**
 * Original code by Claude Duguay for "Java Pro"
 * JSortTable can be found here:
 * <a href="http://www.fawcette.com/javapro/2002_08/magazine/columns/visualcomponents/default_pf.aspx">
 * http://www.fawcette.com/javapro/2002_08/magazine/columns/visualcomponents/default_pf.aspx
 * </a><br/>
 * <br/>
 * Creates a table whose column-headers are used to sort the data.<br/>
 * <br/>
 * Basicly, the only thing this class does is add a mouselistener to the header
 * of the table and respond to mouseclicks on the header by sorting the underlying model.<br/>
 * <br/>
 * - Changed to make the 'sortedColumnIndex' reflect the model's index
 * instead of the 'visible' index. When you dragged a column, the arrow
 * stuck with the 'x-th' column, even though you dragged it to another place.
 * Now the renderer compares the index of the column in the model and it
 * 'sticks' to the column being dragged.
 *
 * @author Claude Duguay
 * @author Vincent Vollers
 */

public class JSortTable
  extends JTable
  implements MouseListener {
  /**
   * The index of the column currently being sorted by.
   */
  protected int sortedColumnIndex = -1;

  /**
   * The order of the sort.
   */
  protected boolean sortedColumnAscending = true;

  /**
   * Creates a new JSortTable with the default models.
   */
  public JSortTable() {
    super();
    initSortHeader();
  }

  /**
   * Creates a new JSortTable with the given SortTableModel.
   */
  public JSortTable(SortTableModel model) {
    super(model);
    initSortHeader();
  }

  /**
   * Creates a new JSortTable with the given SortTableModel and TableColumnModel.
   */
  public JSortTable(SortTableModel model, TableColumnModel colModel) {
    super(model, colModel);
    initSortHeader();
  }

  /**
   * Creates a new JSortTable with the given SortTableModel, TableColumnModel and ListSelectionModel.
   */
  public JSortTable(SortTableModel model,
                    TableColumnModel colModel,
                    ListSelectionModel selModel) {
    super(model, colModel, selModel);
    initSortHeader();
  }

  /**
   * Initializes the header. Adds this class as the mouseListener
   * and setup the headerRenderer.
   */
  protected void initSortHeader() {
    JTableHeader header = getTableHeader();
    header.setDefaultRenderer(new SortHeaderRenderer());
    header.addMouseListener(this);
  }

  /**
   * Returns the index of the currently sorted column.
   *
   * @return the index of the sorted column.
   */
  public int getSortedColumnIndex() {
    return sortedColumnIndex;
  }

  /**
   * Returns the current sort-order.
   *
   * @return true when ascending, false when descending.
   */
  public boolean isSortedColumnAscending() {
    return sortedColumnAscending;
  }

  /**
   * When a user clicks in the header of this table,
   * determine which column was clicked on and if the
   * column is sortable, sort by that column. (swap the
   * sort-order if we already were sorting this column)
   */
  public void mouseClicked(MouseEvent event) {
    TableColumnModel colModel = getColumnModel();

    // get the visible column index at the current mouse coordinates
    int index = colModel.getColumnIndexAtX(event.getX());

    // get the index of that column in the model
    int modelIndex = colModel.getColumn(index).getModelIndex();

    // sort the model by the selected column if that column is sortable
    SortTableModel model = (SortTableModel) getModel();
    if (model.isSortable(modelIndex)) {
      // toggle ascension, if already sorted
      if (sortedColumnIndex == modelIndex) {
        sortedColumnAscending = !sortedColumnAscending;
      }
      sortedColumnIndex = modelIndex;

      model.sortColumn(modelIndex, sortedColumnAscending);
    }
    getTableHeader().repaint();
  }

  /**
   * Dummy method
   */
  public void mouseEntered(MouseEvent event) {
  }

  /**
   * Dummy method
   */
  public void mouseExited(MouseEvent event) {
  }

  /**
   * Dummy method
   */
  public void mouseReleased(MouseEvent event) {
  }

  /**
   * Dummy method
   */
  public void mousePressed(MouseEvent event) {
  }

  public void sort(int column, boolean ascending) {
    SortTableModel model = (SortTableModel) getModel();

    if (model.isSortable(column)) {
      sortedColumnAscending = ascending;
      sortedColumnIndex = column;
      model.sortColumn(column, sortedColumnAscending);
    }
  }

  public void repaintCell(int col, int row) {
    tableChanged(new TableModelEvent(getModel(), row, row, col)); 
    repaint();
  }
}