/*
(C) 2007 Stefan Reich (jazz@drjava.de)
This source file is part of Project Prophecy.
For up-to-date information, see http://www.drjava.de/prophecy

This source file is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation, version 2.1.
*/

package sorttreetable;

import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.table.AbstractTableModel;
import javax.swing.tree.TreePath;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;

/**
 * This is a wrapper class takes a TreeTableModel and implements the table
 * model interface. The implementation is trivial, with all of the event
 * dispatching support provided by the superclass: the AbstractTableModel.<br/>
 *<br/>
 * Original code by Philip Milne and Scott Violet for Sun Microsystems<br/>
 * TreeTable Example 2 can be found here:<br/>
 * <a href="http://java.sun.com/products/jfc/tsc/articles/treetable2/">
 * http://java.sun.com/products/jfc/tsc/articles/treetable2/
 * </a><br/>
 *<br/>
 * - Added wrapper methods for the SortTableModel<br/>
 * - Added methods to get/set the current 'open' nodes 
 *   (used after sorting the tablemodel)<br/>
 * 
 * @author Philip Milne
 * @author Scott Violet
 * @author Vincent Vollers
 */
public class TreeTableModelAdapter extends AbstractTableModel implements SortTableModel {
	protected JTree tree;
	protected AbstractSortTreeTableModel treeTableModel;
	protected JSortTreeTable treeTable;

	/***
	 * Constructs a TreeTableModelAdapter, bridging the given treeTable and the model
	 * 
	 * @param treeTableModel the model to use
	 * @param treeTable the table which is going to use this model
	 */
	public TreeTableModelAdapter(
		AbstractSortTreeTableModel treeTableModel,
		JSortTreeTable treeTable) {
		this.treeTable = treeTable;
		this.tree = treeTable.getTree();
		this.treeTableModel = treeTableModel;

		tree.addTreeExpansionListener(new TreeExpansionListener() {
			// Don't use fireTableRowsInserted() here; the selection model
			// would get updated twice.
			public void treeExpanded(TreeExpansionEvent event) {
				fireTableDataChanged();
			}
			public void treeCollapsed(TreeExpansionEvent event) {
				fireTableDataChanged();
			}
		});

		// Install a TreeModelListener that can update the table when
		// tree changes. We use delayedFireTableDataChanged as we can
		// not be guaranteed the tree will have finished processing
		// the event before us.
		treeTableModel.addTreeModelListener(new TreeModelListener() {
			public void treeNodesChanged(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}

			public void treeNodesInserted(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}

			public void treeNodesRemoved(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}

			public void treeStructureChanged(TreeModelEvent e) {
				delayedFireTableDataChanged();
			}
		});
	}

	/***
	 * TableModel wrapper, passes it through to the model.
	 */
	public int getColumnCount() {
		return treeTableModel.getColumnCount();
	}

	/***
	 * TableModel wrapper, passes it through to the model.
	 */
	public String getColumnName(int column) {
		return treeTableModel.getColumnName(column);
	}

	/***
	 * TableModel wrapper, passes it through to the model.
	 */
	public Class getColumnClass(int column) {
		return treeTableModel.getColumnClass(column);
	}

	/***
	 * TableModel wrapper, passes it through to the model.
	 */
	public int getRowCount() {
		return tree.getRowCount();
	}

	/***
	 * Returns the object (TreeTableNode) on the given row in the tree.
	 */
	protected Object nodeForRow(int row) {
		TreePath treePath = tree.getPathForRow(row);
		if(treePath != null) {
			return treePath.getLastPathComponent();
		} else {
			return null;
		}
	}

	/***
	 * Returns the object on the given row and column.
	 */
	public Object getValueAt(int row, int column) {
		Object j = treeTableModel.getValueAt(nodeForRow(row), column);
		return j;
	}

	/***
	 * TableModel wrapper, passes it through to the model after 
	 * fetching the right TreeTableNode for the given row.
	 */
	public boolean isCellEditable(int row, int column) {
		return treeTableModel.isCellEditable(nodeForRow(row), column);
	}

	/***
	 * TableModel wrapper, passes it through to the model after 
	 * fetching the right TreeTableNode for the given row.
	 */
	public void setValueAt(Object value, int row, int column) {
		treeTableModel.setValueAt(value, nodeForRow(row), column);
	}

	/**
	 * Invokes fireTableDataChanged after all the pending events have been
	 * processed. SwingUtilities.invokeLater is used to handle this.
	 */
	protected void delayedFireTableDataChanged() {
		SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					fireTableDataChanged();
				}
			});
	}
	
	public void updateTreeTable() {
		SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					Vector pathState = getExpandedPaths();
	
					treeTableModel.fireTreeStructureChanged(this, ((TreeTableNode) treeTableModel.getRoot()).getPath(), null, null);

					restoreExpandedPaths(pathState);
			
					treeTable.getTableHeader().repaint();
					
					delayedFireTableDataChanged();
				}
			});
	}

	/***
	 * TableModel wrapper, passes it through to the model.
	 */
	public boolean isSortable(int col) {
		return treeTableModel.isSortable(col);
	}

	/***
	 * Returns the (tree)path to the root of the model.
	 * @return
	 */
	public TreePath getRootPath() {
		TreeTableNode rootNode = (TreeTableNode) treeTableModel.getRoot();
		
		return new TreePath(rootNode.getPath());
	}
	
	/***
	 * Opens the root node.
	 */
	public void expandRoot() {
		tree.expandPath(getRootPath());
	}
	
	/***
	 * Returns a vector of open paths in the tree, can be used to
	 * re-open the paths in a tree after a call to 'treeStructureChanged'
	 * (which causes all open paths to collapse)
	 */
	public Vector getExpandedPaths() {
		Enumeration expanded = 
			tree.getExpandedDescendants(
				getRootPath()
			);
		
		Vector paths = new Vector();
		while(expanded.hasMoreElements()) {
			paths.add(expanded.nextElement());
		}
		
		return paths;
	}
	
	/***
	 * Restores the given open paths on the treeModel.
	 * @param paths a Vector of TreePaths which are going to be opened.
	 */
	public void restoreExpandedPaths(Vector paths) {
		Enumeration e = paths.elements();
		while(e.hasMoreElements()) {
			TreePath path = (TreePath) e.nextElement();
			tree.expandPath(path);
		}
	}
	
	/**
	 * Saves all open paths, sorts the tree, fires a 'treeStructureChanged' 
	 * (which causes Swing to re-draw/cache the tree) and opens the saved paths again.
	 * @see SortTableModel#sortColumn(int, boolean)
	 */
	public void sortColumn(int col, boolean ascending) {
		treeTableModel.sortColumn(col, ascending);
		
		updateTreeTable();
	}
	
	/***
	 * Sorts the model again, according to the previously set index and sortOrder
	 */
	public void sort() {
		int index = treeTable.getSortedColumnIndex();
		boolean ascending = treeTable.isSortedColumnAscending();
		
		if(index != -1) {
			treeTableModel.sortColumn(index, ascending);
		}
		
		updateTreeTable();
	}
	
	/***
	 * Opens all paths in the given node and all nodes below that
	 * @param node
	 */
	public void openAllPaths(TreeTableNode node) {
		tree.expandPath(new TreePath(node.getPath()));
		Enumeration children = node.children();
		while(children.hasMoreElements()) {
			TreeTableNode child = (TreeTableNode) children.nextElement();
			openAllPaths(child);
		}
	}
	
	/***
	 * Opens all paths in the tree (opens all tree-nodes)
	 */
	public void openAllPaths() {
		openAllPaths(
			(TreeTableNode) treeTableModel.getRoot()
		);
	}
	
	/***
	 * TableModel wrapper, passes it through to the model.
	 * @see SortTableModel#getComparator()
	 */
	public SortTableComparator getComparator() {
		return treeTableModel.getComparator();
	}

	/***
	 * TableModel wrapper, passes it through to the model.
	 * @see SortTableModel#setComparator(SortTableComparator)
	 */
	public void setComparator(SortTableComparator sorter) {
		treeTableModel.setComparator(sorter);
	}
}